@@ -3,7 +3,7 @@ require 'sass/script/lexer'
 module Sass
   module Script
     # The parser for SassScript.
-    # It parses a string of code into a tree of {Script::Node}s.
+    # It parses a string of code into a tree of {Script::Tree::Node}s.
     class Parser
       # The line number of the parser's current position.
       #
@@ -12,11 +12,18 @@ module Sass
         @lexer.line
       end
 
+      # The column number of the parser's current position.
+      #
+      # @return [Fixnum]
+      def offset
+        @lexer.offset
+      end
+
       # @param str [String, StringScanner] The source text to parse
       # @param line [Fixnum] The line on which the SassScript appears.
-      #   Used for error reporting
-      # @param offset [Fixnum] The number of characters in on which the SassScript appears.
-      #   Used for error reporting
+      #   Used for error reporting and sourcemap building
+      # @param offset [Fixnum] The character (not byte) offset where the script starts in the line.
+      #   Used for error reporting and sourcemap building
       # @param options [{Symbol => Object}] An options hash;
       #   see {file:SASS_REFERENCE.md#sass_options the Sass options documentation}
       def initialize(str, line, offset, options = {})
@@ -29,13 +36,19 @@ module Sass
       # which signals the end of an interpolated segment,
       # it returns rather than throwing an error.
       #
-      # @return [Script::Node] The root node of the parse tree
+      # @param warn_for_color [Boolean] Whether raw color values passed to
+      #   interoplation should cause a warning.
+      # @return [Script::Tree::Node] The root node of the parse tree
       # @raise [Sass::SyntaxError] if the expression isn't valid SassScript
-      def parse_interpolated
+      def parse_interpolated(warn_for_color = false)
+        # Start two characters back to compensate for #{
+        start_pos = Sass::Source::Position.new(line, offset - 2)
         expr = assert_expr :expr
         assert_tok :end_interpolation
+        expr = Sass::Script::Tree::Interpolation.new(
+          nil, expr, nil, !:wb, !:wa, !:originally_text, warn_for_color)
         expr.options = @options
-        expr
+        node(expr, start_pos)
       rescue Sass::SyntaxError => e
         e.modify_backtrace :line => @lexer.line, :filename => @options[:filename]
         raise e
@@ -43,7 +56,7 @@ module Sass
 
       # Parses a SassScript expression.
       #
-      # @return [Script::Node] The root node of the parse tree
+      # @return [Script::Tree::Node] The root node of the parse tree
       # @raise [Sass::SyntaxError] if the expression isn't valid SassScript
       def parse
         expr = assert_expr :expr
@@ -58,8 +71,8 @@ module Sass
       # Parses a SassScript expression,
       # ending it when it encounters one of the given identifier tokens.
       #
-      # @param [#include?(String)] A set of strings that delimit the expression.
-      # @return [Script::Node] The root node of the parse tree
+      # @param tokens [#include?(String)] A set of strings that delimit the expression.
+      # @return [Script::Tree::Node] The root node of the parse tree
       # @raise [Sass::SyntaxError] if the expression isn't valid SassScript
       def parse_until(tokens)
         @stop_at = tokens
@@ -74,14 +87,17 @@ module Sass
 
       # Parses the argument list for a mixin include.
       #
-      # @return [(Array<Script::Node>, {String => Script::Node}, Script::Node)]
+      # @return [(Array<Script::Tree::Node>,
+      #          {String => Script::Tree::Node},
+      #          Script::Tree::Node,
+      #          Script::Tree::Node)]
       #   The root nodes of the positional arguments, keyword arguments, and
-      #   splat argument. Keyword arguments are in a hash from names to values.
+      #   splat argument(s). Keyword arguments are in a hash from names to values.
       # @raise [Sass::SyntaxError] if the argument list isn't valid SassScript
       def parse_mixin_include_arglist
         args, keywords = [], {}
         if try_tok(:lparen)
-          args, keywords, splat = mixin_arglist || [[], {}]
+          args, keywords, splat, kwarg_splat = mixin_arglist
           assert_tok(:rparen)
         end
         assert_done
@@ -89,7 +105,8 @@ module Sass
         args.each {|a| a.options = @options}
         keywords.each {|k, v| v.options = @options}
         splat.options = @options if splat
-        return args, keywords, splat
+        kwarg_splat.options = @options if kwarg_splat
+        return args, keywords, splat, kwarg_splat
       rescue Sass::SyntaxError => e
         e.modify_backtrace :line => @lexer.line, :filename => @options[:filename]
         raise e
@@ -97,7 +114,7 @@ module Sass
 
       # Parses the argument list for a mixin definition.
       #
-      # @return [(Array<Script::Node>, Script::Node)]
+      # @return [(Array<Script::Tree::Node>, Script::Tree::Node)]
       #   The root nodes of the arguments, and the splat argument.
       # @raise [Sass::SyntaxError] if the argument list isn't valid SassScript
       def parse_mixin_definition_arglist
@@ -117,7 +134,7 @@ module Sass
 
       # Parses the argument list for a function definition.
       #
-      # @return [(Array<Script::Node>, Script::Node)]
+      # @return [(Array<Script::Tree::Node>, Script::Tree::Node)]
       #   The root nodes of the arguments, and the splat argument.
       # @raise [Sass::SyntaxError] if the argument list isn't valid SassScript
       def parse_function_definition_arglist
@@ -138,7 +155,7 @@ module Sass
       # Parse a single string value, possibly containing interpolation.
       # Doesn't assert that the scanner is finished after parsing.
       #
-      # @return [Script::Node] The root node of the parse tree.
+      # @return [Script::Tree::Node] The root node of the parse tree.
       # @raise [Sass::SyntaxError] if the string isn't valid SassScript
       def parse_string
         unless (peek = @lexer.peek) &&
@@ -159,7 +176,7 @@ module Sass
       # Parses a SassScript expression.
       #
       # @overload parse(str, line, offset, filename = nil)
-      # @return [Script::Node] The root node of the parse tree
+      # @return [Script::Tree::Node] The root node of the parse tree
       # @see Parser#initialize
       # @see Parser#parse
       def self.parse(*args)
@@ -186,7 +203,7 @@ module Sass
           PRECEDENCE.each_with_index do |e, i|
             return i if Array(e).include?(op)
           end
-          raise "[BUG] Unknown operator #{op}"
+          raise "[BUG] Unknown operator #{op.inspect}"
         end
 
         # Returns whether or not the given operation is associative.
@@ -205,17 +222,18 @@ module Sass
         def production(name, sub, *ops)
           class_eval <<RUBY, __FILE__, __LINE__ + 1
             def #{name}
-              interp = try_ops_after_interp(#{ops.inspect}, #{name.inspect}) and return interp
+              interp = try_ops_after_interp(#{ops.inspect}, #{name.inspect})
+              return interp if interp
               return unless e = #{sub}
-              while tok = try_tok(#{ops.map {|o| o.inspect}.join(', ')})
+              while tok = try_toks(#{ops.map {|o| o.inspect}.join(', ')})
                 if interp = try_op_before_interp(tok, e)
-                  return interp unless other_interp = try_ops_after_interp(#{ops.inspect}, #{name.inspect}, interp)
+                  other_interp = try_ops_after_interp(#{ops.inspect}, #{name.inspect}, interp)
+                  return interp unless other_interp
                   return other_interp
                 end
 
-                line = @lexer.line
-                e = Operation.new(e, assert_expr(#{sub.inspect}), tok.type)
-                e.line = line
+                e = node(Tree::Operation.new(e, assert_expr(#{sub.inspect}), tok.type),
+                         e.source_range.start_pos)
               end
               e
             end
@@ -226,11 +244,10 @@ RUBY
           class_eval <<RUBY, __FILE__, __LINE__ + 1
             def unary_#{op}
               return #{sub} unless tok = try_tok(:#{op})
-              interp = try_op_before_interp(tok) and return interp
-              line = @lexer.line 
-              op = UnaryOperation.new(assert_expr(:unary_#{op}), :#{op})
-              op.line = line
-              op
+              interp = try_op_before_interp(tok)
+              return interp if interp
+              start_pos = source_position
+              node(Tree::UnaryOperation.new(assert_expr(:unary_#{op}), :#{op}), start_pos)
             end
 RUBY
         end
@@ -238,21 +255,61 @@ RUBY
 
       private
 
+      def source_position
+        Sass::Source::Position.new(line, offset)
+      end
+
+      def range(start_pos, end_pos = source_position)
+        Sass::Source::Range.new(start_pos, end_pos, @options[:filename], @options[:importer])
+      end
+
       # @private
       def lexer_class; Lexer; end
 
+      def map
+        start_pos = source_position
+        e = interpolation
+        return unless e
+        return list e, start_pos unless @lexer.peek && @lexer.peek.type == :colon
+
+        pair = map_pair(e)
+        map = node(Sass::Script::Tree::MapLiteral.new([pair]), start_pos)
+        while try_tok(:comma)
+          pair = map_pair
+          return map unless pair
+          map.pairs << pair
+        end
+        map
+      end
+
+      def map_pair(key = nil)
+        return unless key ||= interpolation
+        assert_tok :colon
+        return key, assert_expr(:interpolation)
+      end
+
       def expr
-        line = @lexer.line
-        return unless e = interpolation
-        list = node(List.new([e], :comma), line)
-        while tok = try_tok(:comma)
-          if interp = try_op_before_interp(tok, list)
-            return interp unless other_interp = try_ops_after_interp([:comma], :expr, interp)
+        start_pos = source_position
+        e = interpolation
+        return unless e
+        list e, start_pos
+      end
+
+      def list(first, start_pos)
+        return first unless @lexer.peek && @lexer.peek.type == :comma
+
+        list = node(Sass::Script::Tree::ListLiteral.new([first], :comma), start_pos)
+        while (tok = try_tok(:comma))
+          element_before_interp = list.elements.length == 1 ? list.elements.first : list
+          if (interp = try_op_before_interp(tok, element_before_interp))
+            other_interp = try_ops_after_interp([:comma], :expr, interp)
+            return interp unless other_interp
             return other_interp
           end
-          list.value << assert_expr(:interpolation)
+          return list unless (e = interpolation)
+          list.elements << e
         end
-        list.value.size == 1 ? list.value.first : list
+        list
       end
 
       production :equals, :interpolation, :single_eq
@@ -260,47 +317,58 @@ RUBY
       def try_op_before_interp(op, prev = nil)
         return unless @lexer.peek && @lexer.peek.type == :begin_interpolation
         wb = @lexer.whitespace?(op)
-        str = Script::String.new(Lexer::OPERATORS_REVERSE[op.type])
-        str.line = @lexer.line
-        interp = Script::Interpolation.new(prev, str, nil, wb, !:wa, :originally_text)
-        interp.line = @lexer.line
+        str = literal_node(Script::Value::String.new(Lexer::OPERATORS_REVERSE[op.type]),
+                           op.source_range)
+        interp = node(
+          Script::Tree::Interpolation.new(prev, str, nil, wb, !:wa, :originally_text),
+          (prev || str).source_range.start_pos)
         interpolation(interp)
       end
 
       def try_ops_after_interp(ops, name, prev = nil)
         return unless @lexer.after_interpolation?
-        return unless op = try_tok(*ops)
-        interp = try_op_before_interp(op, prev) and return interp
+        op = try_toks(*ops)
+        return unless op
+        interp = try_op_before_interp(op, prev)
+        return interp if interp
 
         wa = @lexer.whitespace?
-        str = Script::String.new(Lexer::OPERATORS_REVERSE[op.type])
+        str = literal_node(Script::Value::String.new(Lexer::OPERATORS_REVERSE[op.type]),
+                           op.source_range)
         str.line = @lexer.line
-        interp = Script::Interpolation.new(prev, str, assert_expr(name), !:wb, wa, :originally_text)
-        interp.line = @lexer.line
-        return interp
+        interp = node(
+          Script::Tree::Interpolation.new(prev, str, assert_expr(name), !:wb, wa, :originally_text),
+          (prev || str).source_range.start_pos)
+        interp
       end
 
       def interpolation(first = space)
         e = first
-        while interp = try_tok(:begin_interpolation)
+        while (interp = try_tok(:begin_interpolation))
           wb = @lexer.whitespace?(interp)
-          line = @lexer.line
-          mid = parse_interpolated
+          mid = assert_expr :expr
+          assert_tok :end_interpolation
           wa = @lexer.whitespace?
-          e = Script::Interpolation.new(e, mid, space, wb, wa)
-          e.line = line
+          e = node(
+            Script::Tree::Interpolation.new(e, mid, space, wb, wa),
+            (e || mid).source_range.start_pos)
         end
         e
       end
 
       def space
-        line = @lexer.line
-        return unless e = or_expr
+        start_pos = source_position
+        e = or_expr
+        return unless e
         arr = [e]
-        while e = or_expr
+        while (e = or_expr)
           arr << e
         end
-        arr.size == 1 ? arr.first : node(List.new(arr, :space), line)
+        if arr.size == 1
+          arr.first
+        else
+          node(Sass::Script::Tree::ListLiteral.new(arr, :space), start_pos)
+        end
       end
 
       production :or_expr, :and_expr, :or
@@ -320,17 +388,26 @@ RUBY
         return if @stop_at && @stop_at.include?(@lexer.peek.value)
 
         name = @lexer.next
-        if color = Color::COLOR_NAMES[name.value.downcase]
-          return node(Color.new(color))
+        if (color = Sass::Script::Value::Color::COLOR_NAMES[name.value.downcase])
+          literal_node(Sass::Script::Value::Color.new(color, name.value), name.source_range)
+        elsif name.value == "true"
+          literal_node(Sass::Script::Value::Bool.new(true), name.source_range)
+        elsif name.value == "false"
+          literal_node(Sass::Script::Value::Bool.new(false), name.source_range)
+        elsif name.value == "null"
+          literal_node(Sass::Script::Value::Null.new, name.source_range)
+        else
+          literal_node(Sass::Script::Value::String.new(name.value, :identifier), name.source_range)
         end
-        node(Script::String.new(name.value, :identifier))
       end
 
       def funcall
-        return raw unless tok = try_tok(:funcall)
-        args, keywords, splat = fn_arglist || [[], {}]
+        tok = try_tok(:funcall)
+        return raw unless tok
+        args, keywords, splat, kwarg_splat = fn_arglist
         assert_tok(:rparen)
-        node(Script::Funcall.new(tok.value, args, keywords, splat))
+        node(Script::Tree::Funcall.new(tok.value, args, keywords, splat, kwarg_splat),
+          tok.source_range.start_pos, source_position)
       end
 
       def defn_arglist!(must_have_parens)
@@ -346,15 +423,16 @@ RUBY
         must_have_default = false
         loop do
           c = assert_tok(:const)
-          var = Script::Variable.new(c.value)
+          var = node(Script::Tree::Variable.new(c.value), c.source_range)
           if try_tok(:colon)
             val = assert_expr(:space)
             must_have_default = true
-          elsif must_have_default
-            raise SyntaxError.new("Required argument #{var.inspect} must come before any optional arguments.")
           elsif try_tok(:splat)
             splat = var
             break
+          elsif must_have_default
+            raise SyntaxError.new(
+              "Required argument #{var.inspect} must come before any optional arguments.")
           end
           res << [var, val]
           break unless try_tok(:comma)
@@ -372,88 +450,108 @@ RUBY
       end
 
       def arglist(subexpr, description)
-        return unless e = send(subexpr)
-
         args = []
-        keywords = {}
+        keywords = Sass::Util::NormalizedMap.new
+        e = send(subexpr)
+
+        return [args, keywords] unless e
+
+        splat = nil
         loop do
           if @lexer.peek && @lexer.peek.type == :colon
             name = e
-            @lexer.expected!("comma") unless name.is_a?(Variable)
+            @lexer.expected!("comma") unless name.is_a?(Tree::Variable)
             assert_tok(:colon)
             value = assert_expr(subexpr, description)
 
-            if keywords[name.underscored_name]
+            if keywords[name.name]
               raise SyntaxError.new("Keyword argument \"#{name.to_sass}\" passed more than once")
             end
 
-            keywords[name.underscored_name] = value
+            keywords[name.name] = value
           else
-            if !keywords.empty?
+            if try_tok(:splat)
+              return args, keywords, splat, e if splat
+              splat, e = e, nil
+            elsif splat
+              raise SyntaxError.new("Only keyword arguments may follow variable arguments (...).")
+            elsif !keywords.empty?
               raise SyntaxError.new("Positional arguments must come before keyword arguments.")
             end
 
-            return args, keywords, e if try_tok(:splat)
-            args << e
+            args << e if e
           end
 
-          return args, keywords unless try_tok(:comma)
+          return args, keywords, splat unless try_tok(:comma)
           e = assert_expr(subexpr, description)
         end
       end
 
       def raw
-        return special_fun unless tok = try_tok(:raw)
-        node(Script::String.new(tok.value))
+        tok = try_tok(:raw)
+        return special_fun unless tok
+        literal_node(Script::Value::String.new(tok.value), tok.source_range)
       end
 
       def special_fun
-        return paren unless tok = try_tok(:special_fun)
-        first = node(Script::String.new(tok.value.first))
-        Sass::Util.enum_slice(tok.value[1..-1], 2).inject(first) do |l, (i, r)|
-          Script::Interpolation.new(
-            l, i, r && node(Script::String.new(r)),
-            false, false)
-        end
+        first = try_tok(:special_fun)
+        return paren unless first
+        str = literal_node(first.value, first.source_range)
+        return str unless try_tok(:string_interpolation)
+        mid = parse_interpolated
+        last = assert_expr(:special_fun)
+        node(Tree::Interpolation.new(str, mid, last, false, false),
+            first.source_range.start_pos)
       end
 
       def paren
         return variable unless try_tok(:lparen)
         was_in_parens = @in_parens
         @in_parens = true
-        line = @lexer.line
-        e = expr
+        start_pos = source_position
+        e = map
+        end_pos = source_position
         assert_tok(:rparen)
-        return e || node(List.new([], :space), line)
+        return e || node(Sass::Script::Tree::ListLiteral.new([], nil), start_pos, end_pos)
       ensure
         @in_parens = was_in_parens
       end
 
       def variable
-        return string unless c = try_tok(:const)
-        node(Variable.new(*c.value))
+        start_pos = source_position
+        c = try_tok(:const)
+        return string unless c
+        node(Tree::Variable.new(*c.value), start_pos)
       end
 
       def string
-        return number unless first = try_tok(:string)
-        return first.value unless try_tok(:begin_interpolation)
-        line = @lexer.line
-        mid = parse_interpolated
+        first = try_tok(:string)
+        return number unless first
+        str = literal_node(first.value, first.source_range)
+        return str unless try_tok(:string_interpolation)
+        mid = assert_expr :expr
+        assert_tok :end_interpolation
         last = assert_expr(:string)
-        interp = StringInterpolation.new(first.value, mid, last)
-        interp.line = line
-        interp
+        node(Tree::StringInterpolation.new(str, mid, last), first.source_range.start_pos)
       end
 
       def number
-        return literal unless tok = try_tok(:number)
+        tok = try_tok(:number)
+        return selector unless tok
         num = tok.value
         num.original = num.to_s unless @in_parens
-        num
+        literal_node(num, tok.source_range.start_pos)
+      end
+
+      def selector
+        tok = try_tok(:selector)
+        return literal unless tok
+        node(tok.value, tok.source_range.start_pos)
       end
 
       def literal
-        (t = try_tok(:color, :bool, :null)) && (return t.value)
+        t = try_tok(:color)
+        return literal_node(t.value, t.source_range) if t
       end
 
       # It would be possible to have unified #assert and #try methods,
@@ -464,20 +562,37 @@ RUBY
         :default => "expression (e.g. 1px, bold)",
         :mixin_arglist => "mixin argument",
         :fn_arglist => "function argument",
+        :splat => "...",
+        :special_fun => '")"',
       }
 
       def assert_expr(name, expected = nil)
-        (e = send(name)) && (return e)
+        e = send(name)
+        return e if e
         @lexer.expected!(expected || EXPR_NAMES[name] || EXPR_NAMES[:default])
       end
 
-      def assert_tok(*names)
-        (t = try_tok(*names)) && (return t)
+      def assert_tok(name)
+        # Avoids an array allocation caused by argument globbing in assert_toks.
+        t = try_tok(name)
+        return t if t
+        @lexer.expected!(Lexer::TOKEN_NAMES[name] || name.to_s)
+      end
+
+      def assert_toks(*names)
+        t = try_toks(*names)
+        return t if t
         @lexer.expected!(names.map {|tok| Lexer::TOKEN_NAMES[tok] || tok}.join(" or "))
       end
 
-      def try_tok(*names)
-        peeked =  @lexer.peek
+      def try_tok(name)
+        # Avoids an array allocation caused by argument globbing in the try_toks method.
+        peeked = @lexer.peek
+        peeked && name == peeked.type && @lexer.next
+      end
+
+      def try_toks(*names)
+        peeked = @lexer.peek
         peeked && names.include?(peeked.type) && @lexer.next
       end
 
@@ -486,8 +601,35 @@ RUBY
         @lexer.expected!(EXPR_NAMES[:default])
       end
 
-      def node(node, line = @lexer.line)
-        node.line = line
+      # @overload node(value, source_range)
+      #   @param value [Sass::Script::Value::Base]
+      #   @param source_range [Sass::Source::Range]
+      # @overload node(value, start_pos, end_pos = source_position)
+      #   @param value [Sass::Script::Value::Base]
+      #   @param start_pos [Sass::Source::Position]
+      #   @param end_pos [Sass::Source::Position]
+      def literal_node(value, source_range_or_start_pos, end_pos = source_position)
+        node(Sass::Script::Tree::Literal.new(value), source_range_or_start_pos, end_pos)
+      end
+
+      # @overload node(node, source_range)
+      #   @param node [Sass::Script::Tree::Node]
+      #   @param source_range [Sass::Source::Range]
+      # @overload node(node, start_pos, end_pos = source_position)
+      #   @param node [Sass::Script::Tree::Node]
+      #   @param start_pos [Sass::Source::Position]
+      #   @param end_pos [Sass::Source::Position]
+      def node(node, source_range_or_start_pos, end_pos = source_position)
+        source_range =
+          if source_range_or_start_pos.is_a?(Sass::Source::Range)
+            source_range_or_start_pos
+          else
+            range(source_range_or_start_pos, end_pos)
+          end
+
+        node.line = source_range.start_pos.line
+        node.source_range = source_range
+        node.filename = @options[:filename]
         node
       end
     end
