@@ -1,7 +1,7 @@
 #!/usr/bin/env ruby
 require File.dirname(__FILE__) + '/../test_helper'
 
-class ConversionTest < Test::Unit::TestCase
+class ConversionTest < MiniTest::Test
   def test_basic
     assert_renders <<SASS, <<SCSS
 foo bar
@@ -351,6 +351,23 @@ foo
 SASS
   end
 
+  def test_preserves_triple_slash_comments
+    assert_renders <<SASS, <<SCSS
+/// foo
+/// bar
+foo
+  /// bip bop
+  /// beep boop
+SASS
+/// foo
+/// bar
+foo {
+  /// bip bop
+  /// beep boop
+}
+SCSS
+  end
+
   def test_loud_comments
     assert_renders <<SASS, <<SCSS
 /* foo
@@ -467,6 +484,23 @@ foo
 SASS
   end
 
+  def test_preserves_double_star_comments
+    assert_renders <<SASS, <<SCSS
+/** foo
+ *  bar
+foo
+  /** bip bop
+   *  beep boop
+SASS
+/** foo
+ *  bar */
+foo {
+  /** bip bop
+   *  beep boop */
+}
+SCSS
+  end
+
   def test_loud_comments_with_weird_indentation
     assert_scss_to_sass <<SASS, <<SCSS
 foo
@@ -499,6 +533,29 @@ foo
 SASS
   end
 
+  def test_loud_comment_containing_silent_comment
+    assert_scss_to_sass <<SASS, <<SCSS
+/*
+ *// foo bar
+SASS
+/*
+// foo bar
+*/
+SCSS
+  end
+
+  def test_silent_comment_containing_loud_comment
+    assert_scss_to_sass <<SASS, <<SCSS
+// /*
+//  * foo bar
+//  */
+SASS
+// /*
+//  * foo bar
+//  */
+SCSS
+  end
+
   def test_immediately_preceding_comments
     assert_renders <<SASS, <<SCSS
 /* Foo
@@ -531,6 +588,28 @@ SASS
 SCSS
   end
 
+  def test_immediately_following_comments
+    assert_sass_to_scss <<SCSS, <<SASS
+.foobar {
+  // trailing comment
+  a: 1px;
+}
+SCSS
+.foobar // trailing comment
+  a: 1px
+SASS
+
+    assert_sass_to_scss <<SCSS, <<SASS
+.foobar {
+  // trailing comment
+  a: 1px;
+}
+SCSS
+.foobar  /* trailing comment */
+  a: 1px
+SASS
+  end
+
   def test_debug
     assert_renders <<SASS, <<SCSS
 foo
@@ -544,6 +623,19 @@ foo {
 SCSS
   end
 
+  def test_error
+    assert_renders <<SASS, <<SCSS
+foo
+  @error "oh no!"
+  bar: baz
+SASS
+foo {
+  @error "oh no!";
+  bar: baz;
+}
+SCSS
+  end
+
   def test_directive_without_children
     assert_renders <<SASS, <<SCSS
 foo
@@ -725,6 +817,10 @@ a
 c
   @each $str in foo, bar, baz, bang
     d: $str
+
+c
+  @each $key, $value in (foo: 1, bar: 2, baz: 3)
+    \#{$key}: $value
 SASS
 a {
   @each $number in 1px 2px 3px 4px {
@@ -737,6 +833,12 @@ c {
     d: $str;
   }
 }
+
+c {
+  @each $key, $value in (foo: 1, bar: 2, baz: 3) {
+    \#{$key}: $value;
+  }
+}
 SCSS
   end
 
@@ -977,6 +1079,19 @@ foo {
 SCSS
   end
 
+  def test_mixin_include_with_hyphen_conversion_keyword_arg
+    assert_renders <<SASS, <<SCSS
+foo
+  +foo-bar($a-b_c: val)
+  a: blip
+SASS
+foo {
+  @include foo-bar($a-b_c: val);
+  a: blip;
+}
+SCSS
+  end
+
   def test_argless_function_definition
     assert_renders <<SASS, <<SCSS
 @function foo()
@@ -1199,6 +1314,17 @@ $val: 20;
 SCSS
   end
 
+  def test_media_with_feature
+    assert_sass_to_scss <<SCSS, <<SASS
+@media screen and (-webkit-transform-3d) {
+  a: b;
+}
+SCSS
+@media screen and (-webkit-transform-3d)
+  a: b
+SASS
+  end
+
   def test_supports_with_expressions
     assert_renders <<SASS, <<SCSS
 $query: "(feature1: val)"
@@ -1347,7 +1473,7 @@ div
 SASS
   end
 
-   def test_loud_comment_conversion
+  def test_loud_comment_conversion
     assert_renders(<<SASS, <<SCSS)
 /*! \#{"interpolated"}
 SASS
@@ -1589,6 +1715,28 @@ SASS
 SCSS
   end
 
+  def test_mixin_var_kwargs
+    assert_scss_to_sass <<SASS, <<SCSS
+=foo($a: b, $c: d)
+  a: $a
+  c: $c
+
+.foo
+  +foo($list..., $map...)
+  +foo(pos, $list..., $kwd: val, $map...)
+SASS
+@mixin foo($a: b, $c: d) {
+  a: $a;
+  c: $c;
+}
+
+.foo {
+  @include foo($list..., $map...);
+  @include foo(pos, $list..., $kwd: val, $map...);
+}
+SCSS
+  end
+
   def test_function_var_args
     assert_scss_to_sass <<SASS, <<SCSS
 @function foo($args...)
@@ -1616,6 +1764,141 @@ SASS
 SCSS
   end
 
+  def test_function_var_kwargs
+    assert_scss_to_sass <<SASS, <<SCSS
+@function foo($a: b, $c: d)
+  @return foo
+
+.foo
+  a: foo($list..., $map...)
+  b: foo(pos, $list..., $kwd: val, $map...)
+SASS
+@function foo($a: b, $c: d) {
+  @return foo;
+}
+
+.foo {
+  a: foo($list..., $map...);
+  b: foo(pos, $list..., $kwd: val, $map...);
+}
+SCSS
+  end
+
+  def test_at_root
+    assert_scss_to_sass <<SASS, <<SCSS
+.foo
+  @at-root
+    .bar
+      a: b
+    .baz
+      c: d
+SASS
+.foo {
+  @at-root {
+    .bar {
+      a: b;
+    }
+    .baz {
+      c: d;
+    }
+  }
+}
+SCSS
+  end
+
+  def test_at_root_with_selector
+    assert_scss_to_sass <<SASS, <<SCSS
+.foo
+  @at-root .bar
+    a: b
+SASS
+.foo {
+  @at-root .bar {
+    a: b;
+  }
+}
+SCSS
+  end
+
+  def test_at_root_without
+    assert_scss_to_sass <<SASS, <<SCSS
+.foo
+  @at-root (without: media rule)
+    a: b
+SASS
+.foo {
+  @at-root (without: media rule) {
+    a: b;
+  }
+}
+SCSS
+  end
+
+  def test_at_root_with
+    assert_scss_to_sass <<SASS, <<SCSS
+.foo
+  @at-root (with: media rule)
+    a: b
+SASS
+.foo {
+  @at-root (with: media rule) {
+    a: b;
+  }
+}
+SCSS
+  end
+
+  def test_function_var_kwargs_with_list
+    assert_scss_to_sass <<SASS, <<SCSS
+@function foo($a: b, $c: d)
+  @return $a, $c
+
+.foo
+  a: foo($list..., $map...)
+SASS
+@function foo($a: b, $c: d) {
+  @return $a, $c;
+}
+
+.foo {
+  a: foo($list..., $map...);
+}
+SCSS
+  end
+
+  def test_keyframes
+    assert_renders(<<SASS, <<SCSS)
+@keyframes identifier
+  0%
+    top: 0
+    left: 0
+  30%
+    top: 50px
+  68%, 72%
+    left: 50px
+  100%
+    top: 100px
+    left: 100%
+SASS
+@keyframes identifier {
+  0% {
+    top: 0;
+    left: 0;
+  }
+  30% {
+    top: 50px;
+  }
+  68%, 72% {
+    left: 50px;
+  }
+  100% {
+    top: 100px;
+    left: 100%;
+  }
+}
+SCSS
+  end
+
   ## Regression Tests
 
   def test_list_in_args
@@ -1692,6 +1975,68 @@ foo {
 SCSS
   end
 
+  def test_keyword_arguments
+    assert_renders(<<SASS, <<SCSS, :dasherize => true)
+$foo: foo($dash-ed: 2px)
+SASS
+$foo: foo($dash-ed: 2px);
+SCSS
+    assert_scss_to_sass(<<SASS, <<SCSS, :dasherize => true)
+$foo: foo($dash-ed: 2px)
+SASS
+$foo: foo($dash_ed: 2px);
+SCSS
+    assert_sass_to_scss(<<SCSS, <<SASS, :dasherize => true)
+$foo: foo($dash-ed: 2px);
+SCSS
+$foo: foo($dash_ed: 2px)
+SASS
+    assert_renders(<<SASS, <<SCSS)
+$foo: foo($under_scored: 1px)
+SASS
+$foo: foo($under_scored: 1px);
+SCSS
+    assert_renders(<<SASS, <<SCSS)
+$foo: foo($dash-ed: 2px, $under_scored: 1px)
+SASS
+$foo: foo($dash-ed: 2px, $under_scored: 1px);
+SCSS
+  end
+
+  def test_ambiguous_negation
+    assert_renders(<<SASS, <<SCSS, :indent => '    ')
+foo
+    ok: -$foo
+    comma: 10px, -$foo
+    needs-parens: 10px (-$foo)
+    no-parens: a 50px + 60px b
+SASS
+foo {
+    ok: -$foo;
+    comma: 10px, -$foo;
+    needs-parens: 10px (-$foo);
+    no-parens: a 50px + 60px b;
+}
+SCSS
+  end
+
+  def test_variable_with_global
+    assert_renders(<<SASS, <<SCSS)
+$var: 1
+
+foo
+  $var: 2 !global
+  $var: 3 !global !default
+SASS
+$var: 1;
+
+foo {
+  $var: 2 !global;
+  $var: 3 !global !default;
+}
+SCSS
+  end
+
   private
 
   def assert_sass_to_sass(sass, options = {})
