@@ -9,11 +9,29 @@ require 'pathname'
 
 module Sass::Script::Functions::UserFunctions
   def option(name)
-    Sass::Script::String.new(@options[name.value.to_sym].to_s)
+    Sass::Script::Value::String.new(@options[name.value.to_sym].to_s)
+  end
+
+  def set_a_variable(name, value)
+    environment.set_var(name.value, value)
+    return Sass::Script::Value::Null.new
+  end
+
+  def set_a_global_variable(name, value)
+    environment.set_global_var(name.value, value)
+    return Sass::Script::Value::Null.new
+  end
+
+  def get_a_variable(name)
+    environment.var(name.value) || Sass::Script::Value::String.new("undefined")
   end
 end
 
-class SassEngineTest < Test::Unit::TestCase
+module Sass::Script::Functions
+  include Sass::Script::Functions::UserFunctions
+end
+
+class SassEngineTest < MiniTest::Test
   FAKE_FILE_NAME = __FILE__.gsub(/rb$/,"sass")
   # A map of erroneous Sass documents to the error messages they should produce.
   # The error messages may be arrays;
@@ -54,9 +72,9 @@ MSG
     "$a: 1b <= 2c" => "Incompatible units: 'c' and 'b'.",
     "$a: 1b >= 2c" => "Incompatible units: 'c' and 'b'.",
     "a\n  b: 1b * 2c" => "2b*c isn't a valid CSS value.",
-    "a\n  b: 1b % 2c" => "Cannot modulo by a number with units: 2c.",
-    "$a: 2px + #ccc" => "Cannot add a number with units (2px) to a color (#cccccc).",
-    "$a: #ccc + 2px" => "Cannot add a number with units (2px) to a color (#cccccc).",
+    "a\n  b: 1b % 2c" => "Incompatible units: 'c' and 'b'.",
+    "$a: 2px + #ccc" => "Cannot add a number with units (2px) to a color (#ccc).",
+    "$a: #ccc + 2px" => "Cannot add a number with units (2px) to a color (#ccc).",
     "& a\n  :b c" => ["Base-level rules cannot contain the parent-selector-referencing character '&'.", 1],
     "a\n  :b\n    c" => "Illegal nesting: Only properties may be nested beneath properties.",
     "$a: b\n  :c d\n" => "Illegal nesting: Nothing may be nested beneath variable declarations.",
@@ -128,6 +146,8 @@ MSG
     '@while' => "Invalid while directive '@while': expected expression.",
     '@debug' => "Invalid debug directive '@debug': expected expression.",
     %Q{@debug "a message"\n  "nested message"} => "Illegal nesting: Nothing may be nested beneath debug directives.",
+    '@error' => "Invalid error directive '@error': expected expression.",
+    %Q{@error "a message"\n  "nested message"} => "Illegal nesting: Nothing may be nested beneath error directives.",
     '@warn' => "Invalid warn directive '@warn': expected expression.",
     %Q{@warn "a message"\n  "nested message"} => "Illegal nesting: Nothing may be nested beneath warn directives.",
     "/* foo\n    bar\n  baz" => "Inconsistent indentation: previous line was indented by 4 spaces, but this line was indented by 2 spaces.",
@@ -145,19 +165,19 @@ MSG
     "$var: true\n@while $var\n  @extend .bar\n  $var: false" => ["Extend directives may only be used within rules.", 3],
     "@for $i from 0 to 1\n  @extend .bar" => ["Extend directives may only be used within rules.", 2],
     "@mixin foo\n  @extend .bar\n@include foo" => ["Extend directives may only be used within rules.", 2],
-    "foo\n  &a\n    b: c" => ["Invalid CSS after \"&\": expected \"{\", was \"a\"\n\n\"a\" may only be used at the beginning of a compound selector.", 2],
-    "foo\n  &1\n    b: c" => ["Invalid CSS after \"&\": expected \"{\", was \"1\"\n\n\"1\" may only be used at the beginning of a compound selector.", 2],
     "foo %\n  a: b" => ['Invalid CSS after "foo %": expected placeholder name, was ""', 1],
     "=foo\n  @content error" => "Invalid content directive. Trailing characters found: \"error\".",
     "=foo\n  @content\n    b: c" => "Illegal nesting: Nothing may be nested beneath @content directives.",
     "@content" => '@content may only be used within a mixin.',
     "=simple\n  .simple\n    color: red\n+simple\n  color: blue" => ['Mixin "simple" does not accept a content block.', 4],
     "@import \"foo\" // bar" => "Invalid CSS after \"\"foo\" \": expected media query list, was \"// bar\"",
+    "@at-root\n  a: b" => "Properties are only allowed within rules, directives, mixin includes, or other properties.",
 
     # Regression tests
     "a\n  b:\n    c\n    d" => ["Illegal nesting: Only properties may be nested beneath properties.", 3],
     "& foo\n  bar: baz\n  blat: bang" => ["Base-level rules cannot contain the parent-selector-referencing character '&'.", 1],
     "a\n  b: c\n& foo\n  bar: baz\n  blat: bang" => ["Base-level rules cannot contain the parent-selector-referencing character '&'.", 3],
+    "@" => "Invalid directive: '@'.",
   }
 
   def teardown
@@ -219,42 +239,30 @@ MSG
   end
 
   def test_import_same_name_different_ext
-    assert_warning <<WARNING do
-WARNING: On line 1 of test_import_same_name_different_ext_inline.sass:
-  It's not clear which file to import for '@import "same_name_different_ext"'.
-  Candidates:
-    same_name_different_ext.sass
-    same_name_different_ext.scss
-  For now I'll choose same_name_different_ext.sass.
-  This will be an error in future versions of Sass.
-WARNING
+    assert_raise_message Sass::SyntaxError, <<ERROR do
+It's not clear which file to import for '@import "same_name_different_ext"'.
+Candidates:
+  same_name_different_ext.sass
+  same_name_different_ext.scss
+Please delete or rename all but one of these files.
+ERROR
       options = {:load_paths => [File.dirname(__FILE__) + '/templates/']}
       munge_filename options
-      result = Sass::Engine.new("@import 'same_name_different_ext'", options).render
-      assert_equal(<<CSS, result)
-.foo {
-  ext: sass; }
-CSS
+      Sass::Engine.new("@import 'same_name_different_ext'", options).render
     end
   end
 
   def test_import_same_name_different_partiality
-    assert_warning <<WARNING do
-WARNING: On line 1 of test_import_same_name_different_partiality_inline.sass:
-  It's not clear which file to import for '@import "same_name_different_partiality"'.
-  Candidates:
-    _same_name_different_partiality.scss
-    same_name_different_partiality.scss
-  For now I'll choose _same_name_different_partiality.scss.
-  This will be an error in future versions of Sass.
-WARNING
+    assert_raise_message Sass::SyntaxError, <<ERROR do
+It's not clear which file to import for '@import "same_name_different_partiality"'.
+Candidates:
+  _same_name_different_partiality.scss
+  same_name_different_partiality.scss
+Please delete or rename all but one of these files.
+ERROR
       options = {:load_paths => [File.dirname(__FILE__) + '/templates/']}
       munge_filename options
-      result = Sass::Engine.new("@import 'same_name_different_partiality'", options).render
-      assert_equal(<<CSS, result)
-.foo {
-  partial: yes; }
-CSS
+      Sass::Engine.new("@import 'same_name_different_partiality'", options).render
     end
   end
 
@@ -466,89 +474,25 @@ SASS
     assert_hash_has(err.sass_backtrace[4], :filename => nil, :mixin => nil, :line => 1)
   end
 
-  def test_basic_mixin_loop_exception
-    render <<SASS
-@mixin foo
-  @include foo
-@include foo
+  def test_recursive_mixin
+    assert_equal <<CSS, render(<<SASS)
+.foo .bar .baz {
+  color: blue; }
+.foo .bar .qux {
+  color: red; }
+.foo .zap {
+  color: green; }
+CSS
+@mixin map-to-rule($map-or-color)
+  @if type-of($map-or-color) == map
+    @each $key, $value in $map-or-color
+      .\#{$key}
+        @include map-to-rule($value)
+  @else
+    color: $map-or-color
+
+@include map-to-rule((foo: (bar: (baz: blue, qux: red), zap: green)))
 SASS
-    assert(false, "Exception not raised")
-  rescue Sass::SyntaxError => err
-    assert_equal("An @include loop has been found: foo includes itself", err.message)
-    assert_hash_has(err.sass_backtrace[0], :mixin => "foo", :line => 2)
-  end
-
-  def test_double_mixin_loop_exception
-    render <<SASS
-@mixin foo
-  @include bar
-@mixin bar
-  @include foo
-@include foo
-SASS
-    assert(false, "Exception not raised")
-  rescue Sass::SyntaxError => err
-    assert_equal(<<MESSAGE.rstrip, err.message)
-An @include loop has been found:
-    foo includes bar
-    bar includes foo
-MESSAGE
-    assert_hash_has(err.sass_backtrace[0], :mixin => "bar", :line => 4)
-    assert_hash_has(err.sass_backtrace[1], :mixin => "foo", :line => 2)
-  end
-
-  def test_deep_mixin_loop_exception
-    render <<SASS
-@mixin foo
-  @include bar
-
-@mixin bar
-  @include baz
-
-@mixin baz
-  @include foo
-
-@include foo
-SASS
-    assert(false, "Exception not raised")
-  rescue Sass::SyntaxError => err
-    assert_equal(<<MESSAGE.rstrip, err.message)
-An @include loop has been found:
-    foo includes bar
-    bar includes baz
-    baz includes foo
-MESSAGE
-    assert_hash_has(err.sass_backtrace[0], :mixin => "baz", :line => 8)
-    assert_hash_has(err.sass_backtrace[1], :mixin => "bar", :line => 5)
-    assert_hash_has(err.sass_backtrace[2], :mixin => "foo", :line => 2)
-  end
-
-  def test_mixin_loop_with_content
-    render <<SASS
-=foo
-  @content
-=bar
-  +foo
-    +bar
-+bar
-SASS
-    assert(false, "Exception not raised")
-  rescue Sass::SyntaxError => err
-    assert_equal("An @include loop has been found: bar includes itself", err.message)
-    assert_hash_has(err.sass_backtrace[0], :mixin => "@content", :line => 5)
-  end
-
-  def test_basic_import_loop_exception
-    import = filename_for_test
-    importer = MockImporter.new
-    importer.add_import(import, "@import '#{import}'")
-
-    engine = Sass::Engine.new("@import '#{import}'", :filename => import,
-      :load_paths => [importer])
-
-    assert_raise_message(Sass::SyntaxError, <<ERR.rstrip) {engine.render}
-An @import loop has been found: #{import} imports itself
-ERR
   end
 
   def test_double_import_loop_exception
@@ -589,9 +533,9 @@ ERR
     opts = {:full_exception => true, :line => 362}
     render(("a\n  b: c\n" * 10) + "d\n  e:\n" + ("f\n  g: h\n" * 10), opts)
   rescue Sass::SyntaxError => e
-    assert_equal(<<CSS, Sass::SyntaxError.exception_to_css(e, opts).split("\n")[0..15].join("\n"))
+    assert_equal(<<CSS, Sass::SyntaxError.exception_to_css(e, opts[:line]).split("\n")[0..15].join("\n"))
 /*
-Syntax error: Invalid property: "e:" (no value).
+Error: Invalid property: "e:" (no value).
         on line 383 of test_exception_css_with_offset_inline.sass
 
 378: a
@@ -611,8 +555,7 @@ CSS
   end
 
   def test_exception_css_with_mixins
-    opts = {:full_exception => true}
-    render(<<SASS, opts)
+    render(<<SASS, :full_exception => true)
 =error-mixin($a)
   color: $a * 1em * 1px
 
@@ -623,9 +566,9 @@ CSS
   +outer-mixin(12)
 SASS
   rescue Sass::SyntaxError => e
-    assert_equal(<<CSS, Sass::SyntaxError.exception_to_css(e, opts).split("\n")[0..13].join("\n"))
+    assert_equal(<<CSS, Sass::SyntaxError.exception_to_css(e).split("\n")[0..13].join("\n"))
 /*
-Syntax error: 12em*px isn't a valid CSS value.
+Error: 12em*px isn't a valid CSS value.
         on line 2 of test_exception_css_with_mixins_inline.sass, in `error-mixin'
         from line 5 of test_exception_css_with_mixins_inline.sass, in `outer-mixin'
         from line 8 of test_exception_css_with_mixins_inline.sass
@@ -643,8 +586,7 @@ CSS
   end
 
   def test_cssize_exception_css
-    opts = {:full_exception => true}
-    render(<<SASS, opts)
+    render(<<SASS, :full_exception => true)
 .filler
   stuff: "stuff!"
 
@@ -654,9 +596,9 @@ a: b
   a: b
 SASS
   rescue Sass::SyntaxError => e
-    assert_equal(<<CSS, Sass::SyntaxError.exception_to_css(e, opts).split("\n")[0..11].join("\n"))
+    assert_equal(<<CSS, Sass::SyntaxError.exception_to_css(e).split("\n")[0..11].join("\n"))
 /*
-Syntax error: Properties are only allowed within rules, directives, mixin includes, or other properties.
+Error: Properties are only allowed within rules, directives, mixin includes, or other properties.
         on line 4 of test_cssize_exception_css_inline.sass
 
 1: .filler
@@ -676,12 +618,12 @@ CSS
   end
 
   def test_http_import
-    assert_equal("@import url(http://fonts.googleapis.com/css?family=Droid+Sans);\n",
+    assert_equal("@import \"http://fonts.googleapis.com/css?family=Droid+Sans\";\n",
       render("@import \"http://fonts.googleapis.com/css?family=Droid+Sans\""))
   end
 
   def test_protocol_relative_import
-    assert_equal("@import url(//fonts.googleapis.com/css?family=Droid+Sans);\n",
+    assert_equal("@import \"//fonts.googleapis.com/css?family=Droid+Sans\";\n",
       render("@import \"//fonts.googleapis.com/css?family=Droid+Sans\""))
   end
 
@@ -711,18 +653,18 @@ SASS
 
   def test_sass_import
     sassc_file = sassc_path("importee")
-    assert !File.exists?(sassc_file)
+    assert !File.exist?(sassc_file)
     renders_correctly "import", { :style => :compact, :load_paths => [File.dirname(__FILE__) + "/templates"] }
-    assert File.exists?(sassc_file)
+    assert File.exist?(sassc_file)
   end
 
   def test_sass_pathname_import
     sassc_file = sassc_path("importee")
-    assert !File.exists?(sassc_file)
+    assert !File.exist?(sassc_file)
     renders_correctly("import",
       :style => :compact,
       :load_paths => [Pathname.new(File.dirname(__FILE__) + "/templates")])
-    assert File.exists?(sassc_file)
+    assert File.exist?(sassc_file)
   end
 
   def test_import_from_global_load_paths
@@ -754,23 +696,23 @@ ERR
   end
 
   def test_no_cache
-    assert !File.exists?(sassc_path("importee"))
+    assert !File.exist?(sassc_path("importee"))
     renders_correctly("import", {
         :style => :compact, :cache => false,
         :load_paths => [File.dirname(__FILE__) + "/templates"],
       })
-    assert !File.exists?(sassc_path("importee"))
+    assert !File.exist?(sassc_path("importee"))
   end
 
   def test_import_in_rule
     assert_equal(<<CSS, render(<<SASS, :load_paths => [File.dirname(__FILE__) + '/templates/']))
 .foo #foo {
-  background-color: #bbaaff; }
+  background-color: #baf; }
 
 .bar {
   a: b; }
   .bar #foo {
-    background-color: #bbaaff; }
+    background-color: #baf; }
 CSS
 .foo
   @import partial
@@ -1089,7 +1031,6 @@ SASS
 @-webkit-keyframes warm {
   from {
     color: black; }
-
   to {
     color: red; } }
 CSS
@@ -1193,7 +1134,13 @@ SASS
   end
 
   def test_default_values_for_mixin_arguments
-    assert_equal("white {\n  color: white; }\n\nblack {\n  color: black; }\n", render(<<SASS))
+    assert_equal(<<CSS, render(<<SASS))
+white {
+  color: #FFF; }
+
+black {
+  color: #000; }
+CSS
 =foo($a: #FFF)
   :color $a
 white
@@ -1203,17 +1150,17 @@ black
 SASS
     assert_equal(<<CSS, render(<<SASS))
 one {
-  color: white;
+  color: #fff;
   padding: 1px;
   margin: 4px; }
 
 two {
-  color: white;
+  color: #fff;
   padding: 2px;
   margin: 5px; }
 
 three {
-  color: white;
+  color: #fff;
   padding: 2px;
   margin: 3px; }
 CSS
@@ -1231,17 +1178,17 @@ three
 SASS
     assert_equal(<<CSS, render(<<SASS))
 one {
-  color: white;
+  color: #fff;
   padding: 1px;
   margin: 4px; }
 
 two {
-  color: white;
+  color: #fff;
   padding: 2px;
   margin: 5px; }
 
 three {
-  color: white;
+  color: #fff;
   padding: 2px;
   margin: 3px; }
 CSS
@@ -1427,6 +1374,44 @@ bar
 SASS
   end
 
+  def test_user_defined_function_variable_scope
+    render(<<SASS)
+bar
+  -no-op: set-a-variable(variable, 5)
+  a: $variable
+SASS
+    flunk("Exception not raised for test_user_defined_function_variable_scope")
+  rescue Sass::SyntaxError => e
+    assert_equal('Undefined variable: "$variable".', e.message)
+  end
+
+  def test_user_defined_function_can_change_global_variable
+    assert_equal(<<CSS, render(<<SASS))
+bar {
+  a: 5; }
+CSS
+$variable: 0
+bar
+  $local: 10
+  -no-op: set-a-global-variable(variable, 5)
+  a: $variable
+SASS
+  end
+
+  def test_user_defined_function_cannot_read_local_variable
+    assert_equal(<<CSS, render(<<SASS))
+bar {
+  global: 0;
+  local: undefined; }
+CSS
+$global: 0
+bar
+  $local: 10
+  global: get-a-variable(global)
+  local: get-a-variable(local)
+SASS
+  end
+
   def test_control_directive_in_nested_property
     assert_equal(<<CSS, render(<<SASS))
 foo {
@@ -1540,7 +1525,7 @@ $a: 5
 @while $a != 0
   a-\#{$a}
     blooble: gloop
-  $a: $a - 1
+  $a: $a - 1 !global
 SASS
   end
 
@@ -1607,50 +1592,43 @@ a
 SASS
   end
 
+  def test_destructuring_each
+    assert_equal <<CSS, render(<<SCSS)
+a {
+  foo: 1px;
+  bar: 2px;
+  baz: 3px; }
+
+c {
+  foo: "Value is bar";
+  bar: "Value is baz";
+  bang: "Value is "; }
+CSS
+a
+  @each $name, $number in (foo: 1px, bar: 2px, baz: 3px)
+    \#{$name}: $number
+c
+  @each $key, $value in (foo bar) (bar, baz) bang
+    \#{$key}: "Value is \#{$value}"
+SCSS
+  end
+
   def test_variable_reassignment
     assert_equal(<<CSS, render(<<SASS))
 a {
   b: 1;
   c: 2; }
 CSS
-$a: 1
 a
+  $a: 1
   b: $a
   $a: 2
   c: $a
 SASS
   end
 
-  def test_variable_scope
-    assert_equal(<<CSS, render(<<SASS))
-a {
-  b-1: c;
-  b-2: c;
-  d: 12; }
-
-b {
-  d: 17; }
-CSS
-$i: 12
-a
-  @for $i from 1 through 2
-    b-\#{$i}: c
-  d: $i
-
-=foo
-  $i: 17
-
-b
-  +foo
-  d: $i
-SASS
-  end
-
   def test_hyphen_underscore_insensitive_variables
     assert_equal(<<CSS, render(<<SASS))
-a {
-  b: c; }
-
 d {
   e: 13;
   f: foobar; }
@@ -1658,10 +1636,8 @@ CSS
 $var-hyphen: 12
 $var_under: foo
 
-a
-  $var_hyphen: 1 + $var_hyphen
-  $var-under: $var-under + bar
-  b: c
+$var_hyphen: 1 + $var_hyphen
+$var-under: $var-under + bar
 
 d
   e: $var-hyphen
@@ -1693,7 +1669,7 @@ SASS
   end
 
   def test_argument_error
-    assert_raise(Sass::SyntaxError) { render("a\n  b: hsl(1)") }
+    assert_raises(Sass::SyntaxError) { render("a\n  b: hsl(1)") }
   end
 
   def test_comments_at_the_top_of_a_document
@@ -1818,11 +1794,9 @@ SASS
 
   def test_loud_comment_is_evaluated
     assert_equal <<CSS, render(<<SASS)
-/*!
- * Hue: 327.21649deg */
+/*! Hue: 327.21649deg */
 CSS
-/*!
-  Hue: \#{hue(#f836a0)}
+/*! Hue: \#{hue(#f836a0)}
 SASS
   end
 
@@ -2040,10 +2014,10 @@ $var: "bar"
 SASS
   end
 
-  def test_interpolation_doesnt_deep_unquote_strings
+  def test_interpolation_deep_unquotes_strings
     assert_equal(<<CSS, render(<<SASS))
 .foo {
-  a: "bar" "baz"; }
+  a: bar baz; }
 CSS
 .foo
   a: \#{"bar" "baz"}
@@ -2229,6 +2203,36 @@ CSS
 SASS
   end
 
+  def test_double_media_bubbling_with_surrounding_rules
+    assert_equal <<CSS, render(<<SASS)
+@media (min-width: 0) {
+  a {
+    a: a; }
+
+  b {
+    before: b;
+    after: b; } }
+  @media (min-width: 0) and (max-width: 5000px) {
+    b {
+      x: x; } }
+
+@media (min-width: 0) {
+  c {
+    c: c; } }
+CSS
+@media (min-width: 0)
+  a
+    a: a
+  b
+    before: b
+    @media (max-width: 5000px)
+      x: x
+    after: b
+  c
+    c: c
+SASS
+  end
+
   def test_rule_media_rule_bubbling
     assert_equal <<CSS, render(<<SASS)
 @media bar {
@@ -2255,9 +2259,10 @@ SASS
   @media print {
     .outside {
       color: black; } }
-    @media print and (a: b) {
-      .outside .inside {
-        border: 1px solid black; } }
+  @media print and (a: b) {
+    .outside .inside {
+      border: 1px solid black; } }
+
   .outside .middle {
     display: block; }
 CSS
@@ -2395,8 +2400,76 @@ $val: 20
 SASS
   end
 
+  def test_at_root
+    assert_equal <<CSS, render(<<SASS)
+.bar {
+  a: b; }
+CSS
+.foo
+  @at-root
+    .bar
+      a: b
+SASS
+  end
+
+  def test_at_root_with_selector
+    assert_equal <<CSS, render(<<SASS)
+.bar {
+  a: b; }
+CSS
+.foo
+  @at-root .bar
+    a: b
+SASS
+  end
+
+  def test_at_root_with_query
+    assert_equal <<CSS, render(<<SASS)
+.foo .bar {
+  a: b; }
+CSS
+.foo
+  @media screen
+    @at-root (without: media)
+      .bar
+        a: b
+SASS
+  end
+
+  def test_variable_assignment_with_global
+    assert_no_warning {assert_equal(<<CSS, render(<<SASS))}
+.foo {
+  a: x; }
+
+.bar {
+  b: x; }
+CSS
+$var: 1
+
+.foo
+  $var: x !global
+  a: $var
+
+.bar
+  b: $var
+SASS
+  end
+
   # Regression tests
 
+  def test_list_separator_with_arg_list
+    assert_equal(<<CSS, render(<<SASS))
+.test {
+  separator: comma; }
+CSS
+@mixin arglist-test($args...)
+  separator: list-separator($args)
+
+.test
+  @include arglist-test(this, is, comma, separated)
+SASS
+  end
+
   def test_parent_mixin_in_content_nested
     assert_equal(<<CSS, render(<<SASS))
 a {
@@ -2467,24 +2540,6 @@ body
 SASS
   end
 
-  def test_tricky_mixin_loop_exception
-    render <<SASS
-@mixin foo($a)
-  @if $a
-    @include foo(false)
-    @include foo(true)
-  @else
-    a: b
-
-a
-  @include foo(true)
-SASS
-    assert(false, "Exception not raised")
-  rescue Sass::SyntaxError => err
-    assert_equal("An @include loop has been found: foo includes itself", err.message)
-    assert_hash_has(err.sass_backtrace[0], :mixin => "foo", :line => 3)
-  end
-
   def test_interpolated_comment_in_mixin
     assert_equal <<CSS, render(<<SASS)
 /*! color: red */
@@ -2828,145 +2883,6 @@ CSS
 SASS
   end
 
-  # Encodings
-
-  unless Sass::Util.ruby1_8?
-    def test_encoding_error
-      render("foo\nbar\nb\xFEaz".force_encoding("utf-8"))
-      assert(false, "Expected exception")
-    rescue Sass::SyntaxError => e
-      assert_equal(3, e.sass_line)
-      assert_equal('Invalid UTF-8 character "\xFE"', e.message)
-    end
-
-    def test_ascii_incompatible_encoding_error
-      template = "foo\nbar\nb_z".encode("utf-16le")
-      template[9] = "\xFE".force_encoding("utf-16le")
-      render(template)
-      assert(false, "Expected exception")
-    rescue Sass::SyntaxError => e
-      assert_equal(3, e.sass_line)
-      assert_equal('Invalid UTF-16LE character "\xFE"', e.message)
-    end
-
-    def test_same_charset_as_encoding
-      assert_renders_encoded(<<CSS, <<SASS)
-@charset "UTF-8";
-fóó {
-  a: b; }
-CSS
-@charset "utf-8"
-fóó
-  a: b
-SASS
-    end
-
-    def test_different_charset_than_encoding
-      assert_renders_encoded(<<CSS.force_encoding("IBM866"), <<SASS)
-@charset "IBM866";
-fóó {
-  a: b; }
-CSS
-@charset "ibm866"
-fóó
-  a: b
-SASS
-    end
-
-    def test_different_encoding_than_system
-      assert_renders_encoded(<<CSS.encode("IBM866"), <<SASS.encode("IBM866"))
-@charset "IBM866";
-тАЬ {
-  a: b; }
-CSS
-тАЬ
-  a: b
-SASS
-    end
-
-    def test_multibyte_charset
-      assert_renders_encoded(<<CSS.encode("UTF-16LE"), <<SASS.encode("UTF-16LE").force_encoding("UTF-8"))
-@charset "UTF-16LE";
-fóó {
-  a: b; }
-CSS
-@charset "utf-16le"
-fóó
-  a: b
-SASS
-    end
-
-    def test_multibyte_charset_without_endian_specifier
-      assert_renders_encoded(<<CSS.encode("UTF-32BE"), <<SASS.encode("UTF-32BE").force_encoding("UTF-8"))
-@charset "UTF-32BE";
-fóó {
-  a: b; }
-CSS
-@charset "utf-32"
-fóó
-  a: b
-SASS
-    end
-
-    def test_utf8_bom
-      assert_renders_encoded(<<CSS, <<SASS.force_encoding("BINARY"))
-@charset "UTF-8";
-fóó {
-  a: b; }
-CSS
-\uFEFFfóó
-  a: b
-SASS
-    end
-
-    def test_utf16le_bom
-      assert_renders_encoded(<<CSS.encode("UTF-16LE"), <<SASS.encode("UTF-16LE").force_encoding("BINARY"))
-@charset "UTF-16LE";
-fóó {
-  a: b; }
-CSS
-\uFEFFfóó
-  a: b
-SASS
-    end
-
-    def test_utf32be_bom
-      assert_renders_encoded(<<CSS.encode("UTF-32BE"), <<SASS.encode("UTF-32BE").force_encoding("BINARY"))
-@charset "UTF-32BE";
-fóó {
-  a: b; }
-CSS
-\uFEFFfóó
-  a: b
-SASS
-    end
-
-    # Encoding Regression Test
-
-    def test_multibyte_prop_name
-      assert_equal(<<CSS, render(<<SASS))
-@charset "UTF-8";
-#bar {
-  cölor: blue; }
-CSS
-#bar
-  cölor: blue
-SASS
-    end
-
-    def test_multibyte_and_interpolation
-      assert_equal(<<CSS, render(<<SCSS, :syntax => :scss))
-#bar {
-  background: a 0%; }
-CSS
-#bar {
-  // 
-  background: \#{a} 0%;
-}
-SCSS
-    end
-  end
-
   def test_original_filename_set
     importer = MockImporter.new
     importer.add_import("imported", "div{color:red}")
@@ -2980,16 +2896,10 @@ SCSS
     assert_equal original_filename, importer.engine("imported").options[:original_filename]
   end
 
-  def test_deprecated_PRECISION
-    assert_warning(<<END) {assert_equal 100_000.0, Sass::Script::Number::PRECISION}
-Sass::Script::Number::PRECISION is deprecated and will be removed in a future release. Use Sass::Script::Number.precision_factor instead.
-END
-  end
-
   def test_changing_precision
-    old_precision = Sass::Script::Number.precision
+    old_precision = Sass::Script::Value::Number.precision
     begin
-      Sass::Script::Number.precision = 8
+      Sass::Script::Value::Number.precision = 8
       assert_equal <<CSS, render(<<SASS)
 div {
   maximum: 1.00000001;
@@ -3000,7 +2910,7 @@ div
   too-much: 1.000000001
 SASS
     ensure
-      Sass::Script::Number.precision = old_precision
+      Sass::Script::Value::Number.precision = old_precision
     end
   end
 
@@ -3194,6 +3104,154 @@ SASS
       ], e.sass_backtrace)
   end
 
+  def test_mixin_with_args_and_varargs_passed_no_var_args
+    assert_equal <<CSS, render(<<SASS, :syntax => :scss)
+.foo {
+  a: 1;
+  b: 2;
+  c: 3; }
+CSS
+@mixin three-or-more-args($a, $b, $c, $rest...) {
+  a: $a;
+  b: $b;
+  c: $c;
+}
+
+.foo {
+  @include three-or-more-args($a: 1, $b: 2, $c: 3);
+}
+SASS
+
+  end
+
+  def test_debug_inspects_sass_objects
+    assert_warning(<<END) {render("@debug (a: 1, b: 2)")}
+test_debug_inspects_sass_objects_inline.sass:1 DEBUG: (a: 1, b: 2)
+END
+    assert_warning(<<END) {render("$map: (a: 1, b: 2); @debug $map", :syntax => :scss)}
+test_debug_inspects_sass_objects_inline.scss:1 DEBUG: (a: 1, b: 2)
+END
+  end
+
+  def test_error_throws_sass_objects
+    assert_raise_message(Sass::SyntaxError, "(a: 1, b: 2)") {render("@error (a: 1, b: 2)")}
+    assert_raise_message(Sass::SyntaxError, "(a: 1, b: 2)") do
+      render("$map: (a: 1, b: 2); @error $map", :syntax => :scss)
+    end
+  end
+
+  def test_default_arg_before_splat
+    assert_equal <<CSS, render(<<SASS, :syntax => :scss)
+.foo-positional {
+  a: 1;
+  b: 2;
+  positional-arguments: 3, 4;
+  keyword-arguments: (); }
+
+.foo-keywords {
+  a: true;
+  positional-arguments: ();
+  keyword-arguments: (c: c, d: d); }
+CSS
+@mixin foo($a: true, $b: null, $arguments...) {
+  a: $a;
+  b: $b;
+  positional-arguments: inspect($arguments);
+  keyword-arguments: inspect(keywords($arguments));
+}
+.foo-positional {
+  @include foo(1, 2, 3, 4);
+}
+.foo-keywords {
+  @include foo($c: c, $d: d);
+}
+SASS
+  end
+
+  def test_keyframes
+    assert_equal <<CSS, render(<<SASS)
+@keyframes identifier {
+  0% {
+    top: 0;
+    left: 0; }
+  30% {
+    top: 50px; }
+  68%, 72% {
+    left: 50px; }
+  100% {
+    top: 100px;
+    left: 100%; } }
+CSS
+@keyframes identifier
+  0%
+    top: 0
+    left: 0
+  \#{"30%"}
+    top: 50px
+  68%, 72%
+    left: 50px
+  100%
+    top: 100px
+    left: 100%
+SASS
+  end
+
+  def test_prefixed_keyframes
+    assert_equal <<CSS, render(<<SASS)
+@-moz-keyframes identifier {
+  0% {
+    top: 0;
+    left: 0; }
+  30% {
+    top: 50px; }
+  68%, 72% {
+    left: 50px; }
+  100% {
+    top: 100px;
+    left: 100%; } }
+CSS
+@-moz-keyframes identifier
+  0%
+    top: 0
+    left: 0
+  \#{"30%"}
+    top: 50px
+  68%, 72%
+    left: 50px
+  100%
+    top: 100px
+    left: 100%
+SASS
+  end
+
+  def test_uppercase_keyframes
+    assert_equal <<CSS, render(<<SASS)
+@KEYFRAMES identifier {
+  0% {
+    top: 0;
+    left: 0; }
+  30% {
+    top: 50px; }
+  68%, 72% {
+    left: 50px; }
+  100% {
+    top: 100px;
+    left: 100%; } }
+CSS
+@KEYFRAMES identifier
+  0%
+    top: 0
+    left: 0
+  \#{"30%"}
+    top: 50px
+  68%, 72%
+    left: 50px
+  100%
+    top: 100px
+    left: 100%
+SASS
+  end
+
   private
 
   def assert_hash_has(hash, expected)
