@@ -1,7 +1,7 @@
 #!/usr/bin/env ruby
 require File.dirname(__FILE__) + '/../test_helper'
 
-class ExtendTest < Test::Unit::TestCase
+class ExtendTest < MiniTest::Test
   def test_basic
     assert_equal <<CSS, render(<<SCSS)
 .foo, .bar {
@@ -145,8 +145,8 @@ SCSS
   end
 
   def test_class_unification
-    assert_unification '.foo.bar', '.baz {@extend .foo}', '.foo.bar, .bar.baz' 
-    assert_unification '.foo.baz', '.baz {@extend .foo}', '.baz' 
+    assert_unification '.foo.bar', '.baz {@extend .foo}', '.foo.bar, .bar.baz'
+    assert_unification '.foo.baz', '.baz {@extend .foo}', '.baz'
   end
 
   def test_id_unification
@@ -154,7 +154,7 @@ SCSS
     assert_unification '.foo#baz', '#baz {@extend .foo}', '#baz'
 
     assert_extend_doesnt_match('#bar', '.foo', :failed_to_unify, 2) do
-      assert_unification '.foo#baz', '#bar {@extend .foo}', '.foo#baz'
+      render_unification '.foo#baz', '#bar {@extend .foo}'
     end
   end
 
@@ -180,7 +180,7 @@ SCSS
     assert_unification 'ns|*.foo', '*|* {@extend .foo}', 'ns|*'
 
     assert_extend_doesnt_match('ns2|*', '.foo', :failed_to_unify, 2) do
-      assert_unification 'ns1|*.foo', 'ns2|* {@extend .foo}', 'ns1|*.foo'
+      render_unification 'ns1|*.foo', 'ns2|* {@extend .foo}'
     end
 
     assert_unification 'ns|*.foo', 'ns|* {@extend .foo}', 'ns|*'
@@ -200,7 +200,7 @@ SCSS
     assert_unification 'ns|a.foo', '*|* {@extend .foo}', 'ns|a'
 
     assert_extend_doesnt_match('ns2|*', '.foo', :failed_to_unify, 2) do
-      assert_unification 'ns1|a.foo', 'ns2|* {@extend .foo}', 'ns1|a.foo'
+      render_unification 'ns1|a.foo', 'ns2|* {@extend .foo}'
     end
 
     assert_unification 'ns|a.foo', 'ns|* {@extend .foo}', 'ns|a'
@@ -227,7 +227,7 @@ SCSS
     assert_unification 'ns|*.foo', '*|a {@extend .foo}', 'ns|*.foo, ns|a'
 
     assert_extend_doesnt_match('ns2|a', '.foo', :failed_to_unify, 2) do
-      assert_unification 'ns1|*.foo', 'ns2|a {@extend .foo}', 'ns1|*.foo'
+      render_unification 'ns1|*.foo', 'ns2|a {@extend .foo}'
     end
 
     assert_unification 'ns|*.foo', 'ns|a {@extend .foo}', 'ns|*.foo, ns|a'
@@ -242,7 +242,7 @@ SCSS
     assert_unification '*|a.foo', 'ns|a {@extend .foo}', '*|a.foo, ns|a'
 
     assert_extend_doesnt_match('h1', '.foo', :failed_to_unify, 2) do
-      assert_unification 'a.foo', 'h1 {@extend .foo}', 'a.foo'
+      render_unification 'a.foo', 'h1 {@extend .foo}'
     end
   end
 
@@ -251,7 +251,7 @@ SCSS
     assert_unification 'ns|a.foo', '*|a {@extend .foo}', 'ns|a'
 
     assert_extend_doesnt_match('ns2|a', '.foo', :failed_to_unify, 2) do
-      assert_unification 'ns1|a.foo', 'ns2|a {@extend .foo}', 'ns1|a.foo'
+      render_unification 'ns1|a.foo', 'ns2|a {@extend .foo}'
     end
 
     assert_unification 'ns|a.foo', 'ns|a {@extend .foo}', 'ns|a'
@@ -270,11 +270,11 @@ SCSS
     assert_unification ':foo.baz', '::foo {@extend .baz}', ':foo.baz, :foo::foo'
 
     assert_extend_doesnt_match('::bar', '.baz', :failed_to_unify, 2) do
-      assert_unification '::foo.baz', '::bar {@extend .baz}', '::foo.baz'
+      render_unification '::foo.baz', '::bar {@extend .baz}'
     end
 
     assert_extend_doesnt_match('::foo(2n+1)', '.baz', :failed_to_unify, 2) do
-      assert_unification '::foo.baz', '::foo(2n+1) {@extend .baz}', '::foo.baz'
+      render_unification '::foo.baz', '::foo(2n+1) {@extend .baz}'
     end
 
     assert_unification '::foo.baz', '::foo {@extend .baz}', '::foo'
@@ -310,9 +310,183 @@ SCSS
   end
 
   def test_negation_unification
-    assert_unification ':not(.foo).baz', ':not(.bar) {@extend .baz}', ':not(.foo).baz, :not(.foo):not(.bar)'
-    assert_unification ':not(.foo).baz', ':not(.foo) {@extend .baz}', ':not(.foo)'
-    assert_unification ':not([a=b]).baz', ':not([a = b]) {@extend .baz}', ':not([a=b])'
+    assert_extends ':not(.foo).baz', ':not(.bar) {@extend .baz}', ':not(.foo).baz, :not(.foo):not(.bar)'
+    # Unifying to :not(.foo) here would reduce the specificity of the original selector.
+    assert_extends ':not(.foo).baz', ':not(.foo) {@extend .baz}', ':not(.foo).baz, :not(.foo)'
+  end
+
+  def test_prefixed_pseudoclass_unification
+    assert_unification(
+      ':nth-child(2n+1 of .foo).baz',
+      ':nth-child(2n of .foo) {@extend .baz}',
+      ':nth-child(2n+1 of .foo).baz, :nth-child(2n+1 of .foo):nth-child(2n of .foo)')
+
+    assert_unification(
+      ':nth-child(2n+1 of .foo).baz',
+      ':nth-child(2n+1 of .bar) {@extend .baz}',
+      ':nth-child(2n+1 of .foo).baz, :nth-child(2n+1 of .foo):nth-child(2n+1 of .bar)')
+
+    assert_unification(
+      ':nth-child(2n+1 of .foo).baz',
+      ':nth-child(2n+1 of .foo) {@extend .baz}',
+      ':nth-child(2n+1 of .foo)')
+  end
+
+  def test_extend_into_not
+    assert_extends(':not(.foo)', '.x {@extend .foo}', ':not(.foo):not(.x)')
+    assert_extends(':not(.foo.bar)', '.x {@extend .bar}', ':not(.foo.bar):not(.foo.x)')
+    assert_extends(
+      ':not(.foo.bar, .baz.bar)',
+      '.x {@extend .bar}',
+      ':not(.foo.bar, .foo.x, .baz.bar, .baz.x)')
+  end
+
+  def test_extend_into_mergeable_pseudoclasses
+    assert_extends(':matches(.foo)', '.x {@extend .foo}', ':matches(.foo, .x)')
+    assert_extends(':matches(.foo.bar)', '.x {@extend .bar}', ':matches(.foo.bar, .foo.x)')
+    assert_extends(
+      ':matches(.foo.bar, .baz.bar)',
+      '.x {@extend .bar}',
+      ':matches(.foo.bar, .foo.x, .baz.bar, .baz.x)')
+
+    assert_extends(':-moz-any(.foo)', '.x {@extend .foo}', ':-moz-any(.foo, .x)')
+    assert_extends(':current(.foo)', '.x {@extend .foo}', ':current(.foo, .x)')
+    assert_extends(':has(.foo)', '.x {@extend .foo}', ':has(.foo, .x)')
+    assert_extends(':host(.foo)', '.x {@extend .foo}', ':host(.foo, .x)')
+    assert_extends(':host-context(.foo)', '.x {@extend .foo}', ':host-context(.foo, .x)')
+    assert_extends(':nth-child(n of .foo)', '.x {@extend .foo}', ':nth-child(n of .foo, .x)')
+    assert_extends(
+      ':nth-last-child(n of .foo)',
+      '.x {@extend .foo}',
+      ':nth-last-child(n of .foo, .x)')
+  end
+
+  def test_complex_extend_into_pseudoclass
+    assert_extends(':not(.bar)', '.x .y {@extend .bar}', ':not(.bar):not(.x .y)')
+    assert_extends(':matches(.bar)', '.x .y {@extend .bar}', ':matches(.bar, .x .y)')
+    assert_extends(':current(.bar)', '.x .y {@extend .bar}', ':current(.bar, .x .y)')
+    assert_extends(':has(.bar)', '.x .y {@extend .bar}', ':has(.bar, .x .y)')
+    assert_extends(':host(.bar)', '.x .y {@extend .bar}', ':host(.bar, .x .y)')
+    assert_extends(':host-context(.bar)', '.x .y {@extend .bar}', ':host-context(.bar, .x .y)')
+    assert_extends(
+      ':-moz-any(.bar)',
+      '.x .y {@extend .bar}',
+      ':-moz-any(.bar, .x .y)')
+    assert_extends(
+      ':nth-child(n of .bar)',
+      '.x .y {@extend .bar}',
+      ':nth-child(n of .bar, .x .y)')
+    assert_extends(
+      ':nth-last-child(n of .bar)',
+      '.x .y {@extend .bar}',
+      ':nth-last-child(n of .bar, .x .y)')
+  end
+
+  def test_extend_over_selector_pseudoclass
+    assert_extends(':not(.foo)', '.x {@extend :not(.foo)}', ':not(.foo), .x')
+    assert_extends(
+      ':matches(.foo, .bar)',
+      '.x {@extend :matches(.foo, .bar)}',
+      ':matches(.foo, .bar), .x')
+  end
+
+  def test_matches_within_not
+    assert_extends(
+      ':not(.foo, .bar)',
+      ':matches(.x, .y) {@extend .foo}',
+      ':not(.foo, .x, .y, .bar)')
+  end
+
+  def test_pseudoclasses_merge
+    assert_extends(':matches(.foo)', ':matches(.bar) {@extend .foo}', ':matches(.foo, .bar)')
+    assert_extends(':-moz-any(.foo)', ':-moz-any(.bar) {@extend .foo}', ':-moz-any(.foo, .bar)')
+    assert_extends(':current(.foo)', ':current(.bar) {@extend .foo}', ':current(.foo, .bar)')
+    assert_extends(
+      ':nth-child(n of .foo)',
+      ':nth-child(n of .bar) {@extend .foo}',
+      ':nth-child(n of .foo, .bar)')
+    assert_extends(
+      ':nth-last-child(n of .foo)',
+      ':nth-last-child(n of .bar) {@extend .foo}',
+      ':nth-last-child(n of .foo, .bar)')
+  end
+
+  def test_nesting_pseudoclasses_merge
+    assert_extends(':has(.foo)', ':has(.bar) {@extend .foo}', ':has(.foo, :has(.bar))')
+    assert_extends(':host(.foo)', ':host(.bar) {@extend .foo}', ':host(.foo, :host(.bar))')
+    assert_extends(
+      ':host-context(.foo)',
+      ':host-context(.bar) {@extend .foo}',
+      ':host-context(.foo, :host-context(.bar))')
+  end
+
+  def test_not_unifies_with_unique_values
+    assert_unification('foo', ':not(bar) {@extend foo}', ':not(bar)')
+    assert_unification('#foo', ':not(#bar) {@extend #foo}', ':not(#bar)')
+  end
+
+  def test_not_adds_no_specificity
+    assert_specificity_equals(':not(.foo)', '.foo')
+  end
+
+  def test_matches_has_a_specificity_range
+    # `:matches(.foo, #bar)` has minimum specificity equal to that of `.foo`,
+    # which means `:matches(.foo, #bar) .a` can have less specificity than
+    # `#b.a`. Thus the selector generated by `#b.a` should be preserved.
+    assert_equal <<CSS, render(<<SCSS)
+:matches(.foo, #bar) .a, :matches(.foo, #bar) #b.a {
+  a: b; }
+CSS
+:matches(.foo, #bar) %x {a: b}
+.a {@extend %x}
+#b.a {@extend %x}
+SCSS
+
+    # `:matches(.foo, #bar)` has maximum specificity equal to that of `#bar`,
+    # which means `:matches(.foo, #bar).b` can have greater specificity than `.a
+    # .b`. Thus the selector generated by `:matches(.foo, #bar).b` should be
+    # preserved.
+    assert_equal <<CSS, render(<<SCSS)
+.a .b, .a .b:matches(.foo, #bar) {
+  a: b; }
+CSS
+.a %x {a: b}
+.b {@extend %x}
+.b:matches(.foo, #bar) {@extend %x}
+SCSS
+  end
+
+  def test_extend_into_not_and_normal_extend
+    assert_equal <<CSS, render(<<SCSS)
+.x:not(.y):not(.bar), .foo:not(.y):not(.bar) {
+  a: b; }
+CSS
+.x:not(.y) {a: b}
+.foo {@extend .x}
+.bar {@extend .y}
+SCSS
+  end
+
+  def test_extend_into_matches_and_normal_extend
+    assert_equal <<CSS, render(<<SCSS)
+.x:matches(.y, .bar), .foo:matches(.y, .bar) {
+  a: b; }
+CSS
+.x:matches(.y) {a: b}
+.foo {@extend .x}
+.bar {@extend .y}
+SCSS
+  end
+
+  def test_multilayer_pseudoclass_extend
+    assert_equal <<CSS, render(<<SCSS)
+:matches(.x, .foo, .bar) {
+  a: b; }
+CSS
+:matches(.x) {a: b}
+.foo {@extend .x}
+.bar {@extend .foo}
+SCSS
   end
 
   def test_comma_extendee
@@ -348,7 +522,7 @@ SCSS
 
   def test_long_extendee_requires_all_selectors
     assert_extend_doesnt_match('.baz', '.foo.bar', :not_found, 2) do
-      assert_extends '.foo', '.baz {@extend .foo.bar}', '.foo'
+      render_extends '.foo', '.baz {@extend .foo.bar}'
     end
   end
 
@@ -372,11 +546,11 @@ SCSS
 
   def test_long_extender_aborts_unification
     assert_extend_doesnt_match('h1.baz', '.foo', :failed_to_unify, 2) do
-      assert_extends 'a.foo#bar', 'h1.baz {@extend .foo}', 'a.foo#bar'
+      render_extends 'a.foo#bar', 'h1.baz {@extend .foo}'
     end
 
     assert_extend_doesnt_match('.bang#baz', '.foo', :failed_to_unify, 2) do
-      assert_extends 'a.foo#bar', '.bang#baz {@extend .foo}', 'a.foo#bar'
+      render_extends 'a.foo#bar', '.bang#baz {@extend .foo}'
     end
   end
 
@@ -392,7 +566,7 @@ SCSS
 
   def test_nested_extender_aborts_unification
     assert_extend_doesnt_match('foo bar', '.foo', :failed_to_unify, 2) do
-      assert_extends 'baz.foo', 'foo bar {@extend .foo}', 'baz.foo'
+      render_extends 'baz.foo', 'foo bar {@extend .foo}'
     end
   end
 
@@ -489,7 +663,7 @@ CSS
 SCSS
   end
 
-  def test_nested_extender_with_early_child_selectors_doesnt_subseq_them
+  def test_nested_extender_with_early_child_selector
     assert_equal <<CSS, render(<<SCSS)
 .foo .bar, .foo .bip > .baz {
   a: b; }
@@ -516,7 +690,7 @@ SCSS
   end
 
   def test_nested_extender_with_trailing_child_selector
-    assert_raise(Sass::SyntaxError, "bar > can't extend: invalid selector") do
+    assert_raises(Sass::SyntaxError, "bar > can't extend: invalid selector") do
       render("bar > {@extend .baz}")
     end
   end
@@ -650,10 +824,10 @@ SCSS
 
   def test_basic_extend_loop
     assert_equal <<CSS, render(<<SCSS)
-.bar, .foo {
+.foo, .bar {
   a: b; }
 
-.foo, .bar {
+.bar, .foo {
   c: d; }
 CSS
 .foo {a: b; @extend .bar}
@@ -663,13 +837,13 @@ SCSS
 
   def test_three_level_extend_loop
     assert_equal <<CSS, render(<<SCSS)
-.baz, .bar, .foo {
+.foo, .baz, .bar {
   a: b; }
 
-.foo, .baz, .bar {
+.bar, .foo, .baz {
   c: d; }
 
-.bar, .foo, .baz {
+.baz, .bar, .foo {
   e: f; }
 CSS
 .foo {a: b; @extend .bar}
@@ -692,6 +866,18 @@ CSS
 SCSS
   end
 
+  def test_cross_loop
+    # The first law of extend means the selector should stick around.
+    assert_equal <<CSS, render(<<SCSS)
+.foo.bar, .foo, .bar {
+  a: b; }
+CSS
+.foo.bar {a: b}
+.foo {@extend .bar}
+.bar {@extend .foo}
+SCSS
+  end
+
   def test_multiple_extender_merges_with_superset_selector
     assert_equal <<CSS, render(<<SCSS)
 a.bar.baz, a.foo {
@@ -810,10 +996,7 @@ SCSS
 
   def test_placeholder_selector_as_modifier
     assert_extend_doesnt_match('div', '%foo', :failed_to_unify, 3) do
-      assert_equal <<CSS, render(<<SCSS)
-a.baz.bar {
-  color: blue; }
-CSS
+      render(<<SCSS)
 a%foo.baz {color: blue}
 .bar {@extend %foo}
 div {@extend %foo}
@@ -833,6 +1016,17 @@ $foo: foo;
 SCSS
   end
 
+  def test_placeholder_in_selector_pseudoclass
+    assert_equal <<CSS, render(<<SCSS)
+:matches(.bar, .baz) {
+  color: blue; }
+CSS
+:matches(%foo) {color: blue}
+.bar {@extend %foo}
+.baz {@extend %foo}
+SCSS
+  end
+
   def test_media_in_placeholder_selector
     assert_equal <<CSS, render(<<SCSS)
 .baz {
@@ -844,16 +1038,11 @@ SCSS
   end
 
   def test_extend_out_of_media
-    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
-DEPRECATION WARNING on line 3 of test_extend_out_of_media_inline.scss:
-  @extending an outer selector from within @media is deprecated.
-  You may only @extend selectors within the same directive.
-  This will be an error in Sass 3.3.
-  It can only work once @extend is supported natively in the browser.
-WARN
-.foo {
-  a: b; }
-CSS
+    assert_raise_message(Sass::SyntaxError, <<ERR) {render(<<SCSS)}
+You may not @extend an outer selector from within @media.
+You may only @extend selectors within the same directive.
+From "@extend .foo" on line 3 of test_extend_out_of_media_inline.scss.
+ERR
 .foo {a: b}
 @media screen {
   .bar {@extend .foo}
@@ -862,18 +1051,11 @@ SCSS
   end
 
   def test_extend_out_of_unknown_directive
-    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
-DEPRECATION WARNING on line 3 of test_extend_out_of_unknown_directive_inline.scss:
-  @extending an outer selector from within @flooblehoof is deprecated.
-  You may only @extend selectors within the same directive.
-  This will be an error in Sass 3.3.
-  It can only work once @extend is supported natively in the browser.
-WARN
-.foo {
-  a: b; }
-
-@flooblehoof {}
-CSS
+    assert_raise_message(Sass::SyntaxError, <<ERR) {render(<<SCSS)}
+You may not @extend an outer selector from within @flooblehoof.
+You may only @extend selectors within the same directive.
+From "@extend .foo" on line 3 of test_extend_out_of_unknown_directive_inline.scss.
+ERR
 .foo {a: b}
 @flooblehoof {
   .bar {@extend .foo}
@@ -882,19 +1064,11 @@ SCSS
   end
 
   def test_extend_out_of_nested_directives
-    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
-DEPRECATION WARNING on line 4 of test_extend_out_of_nested_directives_inline.scss:
-  @extending an outer selector from within @flooblehoof is deprecated.
-  You may only @extend selectors within the same directive.
-  This will be an error in Sass 3.3.
-  It can only work once @extend is supported natively in the browser.
-WARN
-@media screen {
-  .foo {
-    a: b; }
-
-  @flooblehoof {} }
-CSS
+    assert_raise_message(Sass::SyntaxError, <<ERR) {render(<<SCSS)}
+You may not @extend an outer selector from within @flooblehoof.
+You may only @extend selectors within the same directive.
+From "@extend .foo" on line 4 of test_extend_out_of_nested_directives_inline.scss.
+ERR
 @media screen {
   .foo {a: b}
   @flooblehoof {
@@ -962,7 +1136,6 @@ SCSS
 @flooblehoof {
   .foo, .bar {
     a: b; } }
-
 @flooblehoof {}
 CSS
 @flooblehoof {.foo {a: b}}
@@ -985,20 +1158,11 @@ SCSS
   end
 
   def test_extend_within_and_without_media
-    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
-DEPRECATION WARNING on line 4 of test_extend_within_and_without_media_inline.scss:
-  @extending an outer selector from within @media is deprecated.
-  You may only @extend selectors within the same directive.
-  This will be an error in Sass 3.3.
-  It can only work once @extend is supported natively in the browser.
-WARN
-.foo {
-  a: b; }
-
-@media screen {
-  .foo, .bar {
-    c: d; } }
-CSS
+    assert_raise_message(Sass::SyntaxError, <<ERR) {render(<<SCSS)}
+You may not @extend an outer selector from within @media.
+You may only @extend selectors within the same directive.
+From "@extend .foo" on line 4 of test_extend_within_and_without_media_inline.scss.
+ERR
 .foo {a: b}
 @media screen {
   .foo {c: d}
@@ -1008,20 +1172,11 @@ SCSS
   end
 
   def test_extend_within_and_without_unknown_directive
-    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
-DEPRECATION WARNING on line 4 of test_extend_within_and_without_unknown_directive_inline.scss:
-  @extending an outer selector from within @flooblehoof is deprecated.
-  You may only @extend selectors within the same directive.
-  This will be an error in Sass 3.3.
-  It can only work once @extend is supported natively in the browser.
-WARN
-.foo {
-  a: b; }
-
-@flooblehoof {
-  .foo, .bar {
-    c: d; } }
-CSS
+    assert_raise_message(Sass::SyntaxError, <<ERR) {render(<<SCSS)}
+You may not @extend an outer selector from within @flooblehoof.
+You may only @extend selectors within the same directive.
+From "@extend .foo" on line 4 of test_extend_within_and_without_unknown_directive_inline.scss.
+ERR
 .foo {a: b}
 @flooblehoof {
   .foo {c: d}
@@ -1031,21 +1186,11 @@ SCSS
   end
 
   def test_extend_within_and_without_nested_directives
-    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
-DEPRECATION WARNING on line 5 of test_extend_within_and_without_nested_directives_inline.scss:
-  @extending an outer selector from within @flooblehoof is deprecated.
-  You may only @extend selectors within the same directive.
-  This will be an error in Sass 3.3.
-  It can only work once @extend is supported natively in the browser.
-WARN
-@media screen {
-  .foo {
-    a: b; }
-
-  @flooblehoof {
-    .foo, .bar {
-      c: d; } } }
-CSS
+    assert_raise_message(Sass::SyntaxError, <<ERR) {render(<<SCSS)}
+You may not @extend an outer selector from within @flooblehoof.
+You may only @extend selectors within the same directive.
+From "@extend .foo" on line 5 of test_extend_within_and_without_nested_directives_inline.scss.
+ERR
 @media screen {
   .foo {a: b}
   @flooblehoof {
@@ -1057,7 +1202,7 @@ SCSS
   end
 
   def test_extend_with_subject_transfers_subject_to_extender
-    assert_equal(<<CSS, render(<<SCSS))
+    silence_warnings {assert_equal(<<CSS, render(<<SCSS))}
 foo bar! baz, foo .bip .bap! baz, .bip foo .bap! baz {
   a: b; }
 CSS
@@ -1065,7 +1210,7 @@ foo bar! baz {a: b}
 .bip .bap {@extend bar}
 SCSS
 
-    assert_equal(<<CSS, render(<<SCSS))
+    silence_warnings {assert_equal(<<CSS, render(<<SCSS))}
 foo.x bar.y! baz.z, foo.x .bip bar.bap! baz.z, .bip foo.x bar.bap! baz.z {
   a: b; }
 CSS
@@ -1075,7 +1220,7 @@ SCSS
   end
 
   def test_extend_with_subject_retains_subject_on_target
-    assert_equal(<<CSS, render(<<SCSS))
+    silence_warnings {assert_equal(<<CSS, render(<<SCSS))}
 .foo! .bar, .foo! .bip .bap, .bip .foo! .bap {
   a: b; }
 CSS
@@ -1085,7 +1230,7 @@ SCSS
   end
 
   def test_extend_with_subject_transfers_subject_to_target
-    assert_equal(<<CSS, render(<<SCSS))
+    silence_warnings {assert_equal(<<CSS, render(<<SCSS))}
 a.foo .bar, .bip a.bap! .bar {
   a: b; }
 CSS
@@ -1095,7 +1240,7 @@ SCSS
   end
 
   def test_extend_with_subject_retains_subject_on_extender
-    assert_equal(<<CSS, render(<<SCSS))
+    silence_warnings {assert_equal(<<CSS, render(<<SCSS))}
 .foo .bar, .foo .bip! .bap, .bip! .foo .bap {
   a: b; }
 CSS
@@ -1105,43 +1250,38 @@ SCSS
   end
 
   def test_extend_with_subject_fails_with_conflicting_subject
-    assert_equal(<<CSS, render(<<SCSS))
+    silence_warnings {assert_equal(<<CSS, render(<<SCSS))}
 x! .bar {
   a: b; }
 CSS
 x! .bar {a: b}
 y! .bap {@extend .bar}
 SCSS
-end
+  end
 
   def test_extend_warns_when_extendee_doesnt_exist
-    assert_warning(<<WARN) {assert_equal("", render(<<SCSS))}
-WARNING on line 1 of test_extend_warns_when_extendee_doesnt_exist_inline.scss: ".foo" failed to @extend ".bar".
-  The selector ".bar" was not found.
-  This will be an error in future releases of Sass.
-  Use "@extend .bar !optional" if the extend should be able to fail.
-WARN
+    assert_raise_message(Sass::SyntaxError, <<ERR) {render(<<SCSS)}
+".foo" failed to @extend ".bar".
+The selector ".bar" was not found.
+Use "@extend .bar !optional" if the extend should be able to fail.
+ERR
 .foo {@extend .bar}
 SCSS
   end
 
   def test_extend_warns_when_extension_fails
-    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
-WARNING on line 2 of test_extend_warns_when_extension_fails_inline.scss: "b.foo" failed to @extend ".bar".
-  No selectors matching ".bar" could be unified with "b.foo".
-  This will be an error in future releases of Sass.
-  Use "@extend .bar !optional" if the extend should be able to fail.
-WARN
-a.bar {
-  a: b; }
-CSS
+    assert_raise_message(Sass::SyntaxError, <<ERR) {render(<<SCSS)}
+"b.foo" failed to @extend ".bar".
+No selectors matching ".bar" could be unified with "b.foo".
+Use "@extend .bar !optional" if the extend should be able to fail.
+ERR
 a.bar {a: b}
 b.foo {@extend .bar}
 SCSS
   end
 
-  def test_extend_does_not_warn_when_one_extension_fails_but_others_dont
-    assert_no_warning {assert_equal(<<CSS, render(<<SCSS))}
+  def test_extend_succeeds_when_one_extension_fails_but_others_dont
+    assert_equal(<<CSS, render(<<SCSS))
 a.bar {
   a: b; }
 
@@ -1154,28 +1294,14 @@ b.foo {@extend .bar}
 SCSS
   end
 
-  def test_extend_does_not_warn_when_one_extension_fails_but_others_dont
-    assert_no_warning {assert_equal(<<CSS, render(<<SCSS))}
-a.bar {
-  a: b; }
-
-.bar, b.foo {
-  c: d; }
-CSS
-a.bar {a: b}
-.bar {c: d}
-b.foo {@extend .bar}
-SCSS
-  end
-
-  def test_optional_extend_does_not_warn_when_extendee_doesnt_exist
-    assert_no_warning {assert_equal("", render(<<SCSS))}
+  def test_optional_extend_succeeds_when_extendee_doesnt_exist
+    assert_equal("", render(<<SCSS))
 .foo {@extend .bar !optional}
 SCSS
   end
 
-  def test_optional_extend_does_not_warn_when_extension_fails
-    assert_no_warning {assert_equal(<<CSS, render(<<SCSS))}
+  def test_optional_extend_succeeds_when_extension_fails
+    assert_equal(<<CSS, render(<<SCSS))
 a.bar {
   a: b; }
 CSS
@@ -1186,6 +1312,109 @@ SCSS
 
   # Regression Tests
 
+  def test_extend_parent_selector_suffix
+    assert_equal <<CSS, render(<<SCSS)
+.a-b, .c {
+  x: y; }
+CSS
+.a {&-b {x: y}}
+.c {@extend .a-b}
+SCSS
+  end
+
+  def test_pseudo_element_superselector
+    # Pseudo-elements shouldn't be removed in superselector calculations.
+    assert_equal <<CSS, render(<<SCSS)
+a#bar, a#bar::fblthp {
+  a: b; }
+CSS
+%x#bar {a: b} // Add an id to make the results have high specificity
+%y, %y::fblthp {@extend %x}
+a {@extend %y}
+SCSS
+
+    # Pseudo-classes can be removed when the second law allows.
+    assert_equal <<CSS, render(<<SCSS)
+a#bar {
+  a: b; }
+CSS
+%x#bar {a: b}
+%y, %y:fblthp {@extend %x}
+a {@extend %y}
+SCSS
+
+    # A few pseudo-elements can be written as pseudo-elements for historical
+    # reasons. See http://www.w3.org/TR/selectors4/#pseudo-elements.
+    %w[first-line first-letter before after].each do |pseudo|
+      assert_equal <<CSS, render(<<SCSS)
+a#bar, a#bar:#{pseudo} {
+  a: b; }
+CSS
+%x#bar {a: b}
+%y, %y:#{pseudo} {@extend %x}
+a {@extend %y}
+SCSS
+    end
+  end
+
+  def test_multiple_source_redundancy_elimination
+    assert_equal <<CSS, render(<<SCSS)
+.test-case, .test-case:active {
+  color: red; }
+
+.test-case:hover {
+  color: green; }
+CSS
+%default-color {color: red}
+%alt-color {color: green}
+
+%default-style {
+  @extend %default-color;
+  &:hover {@extend %alt-color}
+  &:active {@extend %default-color}
+}
+
+.test-case {@extend %default-style}
+SCSS
+  end
+
+  def test_nested_sibling_extend
+    assert_equal <<CSS, render(<<SCSS)
+.parent .bar, .parent .foo {
+  width: 2000px; }
+CSS
+.foo {@extend .bar}
+
+.parent {
+  .bar {
+    width: 2000px;
+  }
+  .foo {
+    @extend .bar
+  }
+}
+SCSS
+  end
+
+  def test_parent_and_sibling_extend
+    assert_equal <<CSS, render(<<SCSS)
+.parent1 .parent2 .child1.child2, .parent2 .parent1 .child1.child2 {
+  c: d; }
+CSS
+%foo %bar%baz {c: d}
+
+.parent1 {
+  @extend %foo;
+  .child1 {@extend %bar}
+}
+
+.parent2 {
+  @extend %foo;
+  .child2 {@extend %baz}
+}
+SCSS
+  end
+
   def test_nested_extend_specificity
     assert_equal <<CSS, render(<<SCSS)
 a :b, a :b:c {
@@ -1339,29 +1568,29 @@ SCSS
 
   def test_extend_cross_branch_redundancy_elimination
     assert_equal <<CSS, render(<<SCSS)
-a c d, b c a d {
+.a .c .d, .b .c .a .d {
   a: b; }
 CSS
-%x c %y {a: b}
-a, b {@extend %x}
-a d {@extend %y}
+%x .c %y {a: b}
+.a, .b {@extend %x}
+.a .d {@extend %y}
 SCSS
 
     assert_equal <<CSS, render(<<SCSS)
-e a c d, a c e d, e b c a d, b c a e d {
+.e .a .c .d, .a .c .e .d, .e .b .c .a .d, .b .c .a .e .d {
   a: b; }
 CSS
-e %z {a: b}
-%x c %y {@extend %z}
-a, b {@extend %x}
-a d {@extend %y}
+.e %z {a: b}
+%x .c %y {@extend %z}
+.a, .b {@extend %x}
+.a .d {@extend %y}
 SCSS
   end
 
   private
 
   def assert_extend_doesnt_match(extender, target, reason, line, syntax = :scss)
-    warn = "\"#{extender}\" failed to @extend \"#{target}\"."
+    message = "\"#{extender}\" failed to @extend \"#{target}\"."
     reason = 
       if reason == :not_found
         "The selector \"#{target}\" was not found."
@@ -1369,15 +1598,14 @@ SCSS
         "No selectors matching \"#{target}\" could be unified with \"#{extender}\"."
       end
 
-    assert_warning(<<WARNING) {yield}
-WARNING on line #{line} of #{filename_for_test syntax}: #{warn}
-  #{reason}
-  This will be an error in future releases of Sass.
-  Use "@extend #{target} !optional" if the extend should be able to fail.
-WARNING
+    assert_raise_message(Sass::SyntaxError, <<ERR) {yield}
+#{message}
+#{reason}
+Use "@extend #{target} !optional" if the extend should be able to fail.
+ERR
   end
 
-  def assert_unification(selector, extension, unified)
+  def assert_unification(selector, extension, unified, nested = true)
     # Do some trickery so the first law of extend doesn't get in our way.
     assert_extends(
       "%-a #{selector}",
@@ -1385,11 +1613,41 @@ WARNING
       unified.split(', ').map {|s| "-a #{s}"}.join(', '))
   end
 
-  def assert_extends(selector, extension, result)
+  def assert_specificity_equals(sel1, sel2)
+    assert_specificity_gte(sel1, sel2)
+    assert_specificity_gte(sel2, sel1)
+  end
+
+  def assert_specificity_gte(sel1, sel2)
     assert_equal <<CSS, render(<<SCSS)
+#{sel1} .a {
+  a: b; }
+CSS
+#{sel1} %-a {a: b}
+.a {@extend %-a}
+#{sel2}.a {@extend %-a}
+SCSS
+  end
+
+  def render_unification(selector, extension)
+    render_extends(
+      "%-a #{selector}",
+      extension + " -a {@extend %-a}")
+  end
+
+  def assert_extends(selector, extension, result)
+    assert_equal <<CSS, render_extends(selector, extension)
 #{result} {
   a: b; }
 CSS
+  end
+
+  def assert_extends_to_nothing(selector, extension)
+    assert_equal '', render_extends(selector, extension)
+  end
+
+  def render_extends(selector, extension)
+    render(<<SCSS)
 #{selector} {a: b}
 #{extension}
 SCSS
