@@ -6,7 +6,7 @@ require 'sass/scss/css_parser'
 # These tests just test the parsing of CSS
 # (both standard and any hacks we intend to support).
 # Tests of SCSS-specific behavior go in scss_test.rb.
-class ScssCssTest < Test::Unit::TestCase
+class ScssCssTest < MiniTest::Test
   include ScssTestHelper
 
   def test_basic_scss
@@ -94,7 +94,7 @@ foo {a /*: b; c */: d}
 SCSS
   end
 
-  def test_crazy_comments 
+  def test_crazy_comments
     # http://www.w3.org/Style/CSS/Test/CSS2.1/current/xhtml1/t040109-c17-comments-00-b.xht
     assert_equal <<CSS, render(<<SCSS)
 /* This is a CSS comment. */
@@ -499,19 +499,40 @@ SCSS
 
   def test_import_directive
     assert_parses '@import "foo.css";'
-    assert_parses "@import 'foo.css';"
     assert_parses '@import url("foo.css");'
     assert_parses "@import url('foo.css');"
     assert_parses '@import url(foo.css);'
+
+    assert_equal <<CSS, render(<<SCSS)
+@import "foo.css";
+CSS
+@import 'foo.css';
+SCSS
   end
 
-  def test_import_directive_with_media
+  def test_import_directive_with_backslash_newline
+    assert_equal <<CSS, render(<<SCSS)
+@import "foobar.css";
+CSS
+@import "foo\\
+bar.css";
+SCSS
+  end
+
+  def test_string_import_directive_with_media
     assert_parses '@import "foo.css" screen;'
     assert_parses '@import "foo.css" screen, print;'
     assert_parses '@import "foo.css" screen, print and (foo: 0);'
     assert_parses '@import "foo.css" screen, only print, screen and (foo: 0);'
   end
 
+  def test_url_import_directive_with_media
+    assert_parses '@import url("foo.css") screen;'
+    assert_parses '@import url("foo.css") screen, print;'
+    assert_parses '@import url("foo.css") screen, print and (foo: 0);'
+    assert_parses '@import url("foo.css") screen, only print, screen and (foo: 0);'
+  end
+
   def test_page_directive
     assert_parses <<SCSS
 @page {
@@ -566,7 +587,6 @@ SCSS
     assert_parses <<SCSS
 @foo bar {
   a: b; }
-
 @bar baz {
   c: d; }
 SCSS
@@ -586,7 +606,6 @@ SCSS
     assert_equal <<CSS, render(<<SCSS)
 @foo {
   a: b; }
-
 @bar {
   a: b; }
 CSS
@@ -645,18 +664,78 @@ CSS
     a: b;
   }
 }
+SCSS
+  end
+
+  def test_keyframes
+    assert_equal <<CSS, render(<<SCSS)
+@keyframes identifier {
+  0% {
+    top: 0;
+    left: 0; }
+  30% {
+    top: 50px; }
+  68%, 72% {
+    left: 50px; }
+  100% {
+    top: 100px;
+    left: 100%; } }
+CSS
+@keyframes identifier {
+  0% {top: 0; left: 0}
+  30% {top: 50px}
+  68%, 72% {left: 50px}
+  100% {top: 100px; left: 100%}
+}
+SCSS
+  end
+
+  def test_keyframes_with_empty_rules
+    assert_equal <<CSS, render(<<SCSS)
+@keyframes identifier {
+  50% {
+    background-color: black; } }
+CSS
+@keyframes identifier {
+  0% {}
+  50% {background-color: black}
+  100% {}
+}
+SCSS
+  end
+
+  def test_keyframes_with_custom_identifiers
+    assert_equal <<CSS, render(<<SCSS)
+@-skrollr-keyframes identifier {
+  center-top {
+    left: 100%; }
+  top-bottom {
+    left: 0%; } }
+CSS
+@-skrollr-keyframes identifier {
+  center-top {
+    left: 100%;
+  }
+
+  top-bottom {
+    left: 0%;
+  }
+}
+
 SCSS
   end
 
   ## Selectors
 
   # Taken from http://dev.w3.org/csswg/selectors4/#overview
-  def test_summarized_selectors
+  def test_summarized_selectors_with_element
     assert_selector_parses('*')
     assert_selector_parses('E')
     assert_selector_parses('E:not(s)')
     assert_selector_parses('E:not(s1, s2)')
     assert_selector_parses('E:matches(s1, s2)')
+    assert_selector_parses('E:has(s1, s2)')
+    assert_selector_parses('E:has(> s1, > s2)')
     assert_selector_parses('E.warning')
     assert_selector_parses('E#myid')
     assert_selector_parses('E[foo]')
@@ -707,28 +786,32 @@ SCSS
     assert_selector_parses('E:last-of-type')
     assert_selector_parses('E:nth-last-of-type(n)')
     assert_selector_parses('E:only-of-type')
-    assert_selector_parses('E:nth-match(n of selector)')
-    assert_selector_parses('E:nth-last-match(n of selector)')
-    assert_selector_parses('E:column(selector)')
-    assert_selector_parses('E:nth-column(n)')
-    assert_selector_parses('E:nth-last-column(n)')
+    assert_selector_parses('E:nth-child(n of selector)')
+    assert_selector_parses('E:nth-last-child(n of selector)')
+    assert_selector_parses('E:nth-child(n)')
+    assert_selector_parses('E:nth-last-child(n)')
     assert_selector_parses('E F')
     assert_selector_parses('E > F')
     assert_selector_parses('E + F')
     assert_selector_parses('E ~ F')
     assert_selector_parses('E /foo/ F')
-    assert_selector_parses('E! > F')
+    silence_warnings {assert_selector_parses('E! > F')}
 
     assert_selector_parses('E /ns|foo/ F')
-    assert_selector_parses('E /*|foo/ F')
+
+    # From http://dev.w3.org/csswg/css-scoping-1/
+    assert_selector_parses('E:host(s)')
+    assert_selector_parses('E:host-context(s)')
   end
 
   # Taken from http://dev.w3.org/csswg/selectors4/#overview, but without element
   # names.
-  def test_summarized_selectors
+  def test_more_summarized_selectors
     assert_selector_parses(':not(s)')
     assert_selector_parses(':not(s1, s2)')
     assert_selector_parses(':matches(s1, s2)')
+    assert_selector_parses(':has(s1, s2)')
+    assert_selector_parses(':has(> s1, > s2)')
     assert_selector_parses('.warning')
     assert_selector_parses('#myid')
     assert_selector_parses('[foo]')
@@ -779,11 +862,14 @@ SCSS
     assert_selector_parses(':last-of-type')
     assert_selector_parses(':nth-last-of-type(n)')
     assert_selector_parses(':only-of-type')
-    assert_selector_parses(':nth-match(n of selector)')
-    assert_selector_parses(':nth-last-match(n of selector)')
-    assert_selector_parses(':column(selector)')
-    assert_selector_parses(':nth-column(n)')
-    assert_selector_parses(':nth-last-column(n)')
+    assert_selector_parses(':nth-child(n of selector)')
+    assert_selector_parses(':nth-last-child(n of selector)')
+    assert_selector_parses(':nth-child(n)')
+    assert_selector_parses(':nth-last-child(n)')
+
+    # From http://dev.w3.org/csswg/css-scoping-1/
+    assert_selector_parses(':host(s)')
+    assert_selector_parses(':host-context(s)')
   end
 
   def test_attribute_selectors_with_identifiers
@@ -823,15 +909,18 @@ SCSS
   def test_selectors_containing_selectors
     assert_selector_can_contain_selectors(':not(<sel>)')
     assert_selector_can_contain_selectors(':current(<sel>)')
-    assert_selector_can_contain_selectors(':nth-match(n of <sel>)')
-    assert_selector_can_contain_selectors(':nth-last-match(n of <sel>)')
-    assert_selector_can_contain_selectors(':column(<sel>)')
+    assert_selector_can_contain_selectors(':nth-child(n of <sel>)')
+    assert_selector_can_contain_selectors(':nth-last-child(n of <sel>)')
     assert_selector_can_contain_selectors(':-moz-any(<sel>)')
+    assert_selector_can_contain_selectors(':has(<sel>)')
+    assert_selector_can_contain_selectors(':has(+ <sel>)')
+    assert_selector_can_contain_selectors(':host(<sel>)')
+    assert_selector_can_contain_selectors(':host-context(<sel>)')
   end
 
   def assert_selector_can_contain_selectors(sel)
     try = lambda {|subsel| assert_selector_parses(sel.gsub('<sel>', subsel))}
-    
+
     try['foo|bar']
     try['*|bar']
 
@@ -883,11 +972,11 @@ SCSS
   end
 
   def test_expression_fallback_selectors
-    assert_selector_parses('0%')
-    assert_selector_parses('60%')
-    assert_selector_parses('100%')
-    assert_selector_parses('12px')
-    assert_selector_parses('"foo"')
+    assert_directive_parses('0%')
+    assert_directive_parses('60%')
+    assert_directive_parses('100%')
+    assert_directive_parses('12px')
+    assert_directive_parses('"foo"')
   end
 
   def test_functional_pseudo_selectors
@@ -924,6 +1013,29 @@ SCSS
     assert_equal "E + F {\n  a: b; }\n", render("E+F { a: b;} ")
   end
 
+  def test_subject_selector_deprecation
+    assert_warning(<<WARNING) {render(".foo .bar! .baz {a: b}")}
+DEPRECATION WARNING on line 1, column 1:
+The subject selector operator "!" is deprecated and will be removed in a future release.
+This operator has been replaced by ":has()" in the CSS spec.
+For example: .foo .bar:has(.baz)
+WARNING
+
+    assert_warning(<<WARNING) {render(".foo .bar! > .baz {a: b}")}
+DEPRECATION WARNING on line 1, column 1:
+The subject selector operator "!" is deprecated and will be removed in a future release.
+This operator has been replaced by ":has()" in the CSS spec.
+For example: .foo .bar:has(> .baz)
+WARNING
+
+    assert_warning(<<WARNING) {render(".foo .bar! {a: b}")}
+DEPRECATION WARNING on line 1, column 1:
+The subject selector operator "!" is deprecated and will be removed in a future release.
+This operator has been replaced by ":has()" in the CSS spec.
+For example: .foo .bar
+WARNING
+  end
+
   ## Errors
 
   def test_invalid_directives
@@ -1009,8 +1121,41 @@ SCSS
     assert_equal 1, e.sass_line
   end
 
+  def test_newline_in_property_value
+    assert_equal(<<CSS, render(<<SCSS))
+.foo {
+  bar: "bazbang"; }
+CSS
+.foo {
+  bar: "baz\\
+bang";
+}
+SCSS
+  end
+
   ## Regressions
 
+  def test_very_long_comment_doesnt_take_forever
+    string = 'asdf' * (100000)
+    assert_equal(<<CSS, render(<<SCSS))
+/*
+  #{string}
+*/
+CSS
+/*
+  #{string}
+*/
+SCSS
+  end
+
+  def test_long_unclosed_comment_doesnt_take_forever
+    assert_raise_message(Sass::SyntaxError,
+      'Invalid CSS after "/*": expected "/", was "//*************..."') {render(<<SCSS)}
+/*
+//**************************************************************************
+SCSS
+  end
+
   def test_double_space_string
     assert_equal(<<CSS, render(<<SCSS))
 .a {
@@ -1083,10 +1228,22 @@ SCSS
 #{selector} {
   a: b; }
 SCSS
+
+    assert_parses <<SCSS
+:not(#{selector}) {
+  a: b; }
+SCSS
+  end
+
+  def assert_directive_parses(param)
+    assert_parses <<SCSS
+@keyframes #{param} {
+  a: b; }
+SCSS
   end
 
   def render(scss, options = {})
-    tree = Sass::SCSS::CssParser.new(scss, options[:filename]).parse
+    tree = Sass::SCSS::CssParser.new(scss, options[:filename], nil).parse
     tree.options = Sass::Engine::DEFAULT_OPTIONS.merge(options)
     tree.render
   end
